# Copyright 2023-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Input/output operations from circuits and simulations."""

from __future__ import annotations
from pathlib import Path
import logging

import bluepy
import numpy as np

from bluecellulab.circuit.node_id import CellId

logger = logging.getLogger(__name__)


def parse_outdat(path: str | Path) -> dict[CellId, np.ndarray]:
    """Parse the replay spiketrains in a out.dat formatted file pointed to by
    path."""
    spikes = bluepy.impl.spike_report.SpikeReport.load(path).get()
    # convert Series to DataFrame with 2 columns for `groupby` operation
    spike_df = spikes.to_frame().reset_index()
    if (spike_df["t"] < 0).any():
        logger.warning('Found negative spike times in out.dat ! '
                       'Clipping them to 0')
        spike_df["t"].clip(lower=0., inplace=True)

    outdat = spike_df.groupby("gid")["t"].apply(np.array)
    # convert outdat's index from int to CellId
    outdat.index = [CellId("", gid) for gid in outdat.index]
    return outdat.to_dict()



# Copyright 2023-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interfacing the cell object with SONATA properties."""

from functools import partialmethod
from typing import Any
from bluecellulab.circuit import CircuitAccess
from bluecellulab.circuit.node_id import CellId
from bluecellulab.exceptions import MissingSonataPropertyError


class SonataProxy:
    """Sonata proxy to be used by Cell objects."""

    def __init__(self, cell_id: CellId, circuit_access: CircuitAccess) -> None:
        self.cell_id = cell_id
        self.circuit_access = circuit_access

    def get_property(self, property_name: str) -> Any:
        """Get a property of a cell."""
        if property_name in self.circuit_access.available_cell_properties:
            return self.circuit_access.get_cell_properties(self.cell_id, property_name)
        raise MissingSonataPropertyError(f"{property_name} property is not available.")

    get_input_resistance = partialmethod(get_property, "@dynamics:input_resistance")



# Copyright 2023-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Layer to abstract the circuit access functionality from the rest of
bluecellulab."""

from __future__ import annotations
from collections import defaultdict
from typing import Any, Optional, Protocol


import pandas as pd
from pydantic.dataclasses import dataclass

from bluecellulab.circuit import CellId
from bluecellulab.circuit.config import SimulationConfig


@dataclass(config=dict(extra="forbid"))
class EmodelProperties:
    threshold_current: float
    holding_current: float
    AIS_scaler: Optional[float] = 1.0
    soma_scaler: Optional[float] = 1.0


def get_synapse_connection_parameters(
        circuit_access: CircuitAccess, pre_cell: CellId, post_cell: CellId) -> dict:
    """Apply connection blocks in order for pre_gid, post_gid to determine a
    final connection override for this pair (pre_gid, post_gid)."""
    parameters: defaultdict[str, Any] = defaultdict(list)
    parameters['add_synapse'] = True

    for entry in circuit_access.config.connection_entries():
        src_matches = circuit_access.target_contains_cell(entry.source, pre_cell)
        dest_matches = circuit_access.target_contains_cell(entry.target, post_cell)

        if src_matches and dest_matches:
            # whatever specified in this block, is applied to gid
            apply_parameters = True

            if entry.delay is not None:
                parameters['DelayWeights'].append((entry.delay, entry.weight))
                apply_parameters = False

            if apply_parameters:
                if entry.weight is not None:
                    parameters['Weight'] = entry.weight
                if entry.spont_minis is not None:
                    parameters['SpontMinis'] = entry.spont_minis
                if entry.synapse_configure is not None:
                    # collect list of applicable configure blocks to be
                    # applied with a "hoc exec" statement
                    parameters['SynapseConfigure'].append(entry.synapse_configure)
                if entry.mod_override is not None:
                    parameters['ModOverride'] = entry.mod_override
    return parameters


class CircuitAccess(Protocol):
    """Protocol that defines the circuit access layer."""

    config: SimulationConfig

    @property
    def available_cell_properties(self) -> set:
        raise NotImplementedError

    def get_emodel_properties(self, cell_id: CellId) -> Optional[EmodelProperties]:
        raise NotImplementedError

    def get_template_format(self) -> Optional[str]:
        raise NotImplementedError

    def get_cell_properties(
        self, cell_id: CellId, properties: list[str] | str
    ) -> pd.Series:
        raise NotImplementedError

    def extract_synapses(
        self, cell_id: CellId, projections: Optional[list[str] | str]
    ) -> pd.DataFrame:
        raise NotImplementedError

    def target_contains_cell(self, target: str, cell_id: CellId) -> bool:
        raise NotImplementedError

    def is_valid_group(self, group: str) -> bool:
        raise NotImplementedError

    def get_target_cell_ids(self, target: str) -> set[CellId]:
        raise NotImplementedError

    def fetch_cell_info(self, cell_id: CellId) -> pd.Series:
        raise NotImplementedError

    def fetch_mini_frequencies(self, cell_id: CellId) -> tuple[float | None, float | None]:
        raise NotImplementedError

    @property
    def node_properties_available(self) -> bool:
        raise NotImplementedError

    def get_gids_of_mtypes(self, mtypes: list[str]) -> set[CellId]:
        raise NotImplementedError

    def get_cell_ids_of_targets(self, targets: list[str]) -> set[CellId]:
        raise NotImplementedError

    def morph_filepath(self, cell_id: CellId) -> str:
        raise NotImplementedError

    def emodel_path(self, cell_id: CellId) -> str:
        raise NotImplementedError



# Copyright 2023-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module that interacts with the simulation config."""

from __future__ import annotations
from typing import Optional, Protocol


from bluecellulab.circuit.config.sections import Conditions, ConnectionOverrides
from bluecellulab.stimulus.circuit_stimulus_definitions import Stimulus


class SimulationConfig(Protocol):
    """Protocol that defines the simulation config layer."""

    def get_all_stimuli_entries(self) -> list[Stimulus]:
        raise NotImplementedError

    def get_all_projection_names(self) -> list[str]:
        raise NotImplementedError

    def condition_parameters(self) -> Conditions:
        raise NotImplementedError

    def connection_entries(self) -> list[ConnectionOverrides]:
        raise NotImplementedError

    @property
    def base_seed(self) -> int:
        raise NotImplementedError

    @property
    def synapse_seed(self) -> int:
        raise NotImplementedError

    @property
    def ionchannel_seed(self) -> int:
        raise NotImplementedError

    @property
    def stimulus_seed(self) -> int:
        raise NotImplementedError

    @property
    def minis_seed(self) -> int:
        raise NotImplementedError

    @property
    def rng_mode(self) -> str:
        raise NotImplementedError

    @property
    def spike_threshold(self) -> float:
        raise NotImplementedError

    @property
    def spike_location(self) -> str:
        raise NotImplementedError

    @property
    def duration(self) -> Optional[float]:
        raise NotImplementedError

    @property
    def dt(self) -> float:
        raise NotImplementedError

    @property
    def forward_skip(self) -> Optional[float]:
        raise NotImplementedError

    @property
    def celsius(self) -> float:
        raise NotImplementedError

    @property
    def v_init(self) -> float:
        raise NotImplementedError

    @property
    def output_root_path(self) -> str:
        raise NotImplementedError

    @property
    def extracellular_calcium(self) -> Optional[float]:
        raise NotImplementedError

    def add_connection_override(
        self,
        connection_override: ConnectionOverrides
    ) -> None:
        raise NotImplementedError



# Copyright 2023-2024 Blue Brain Project / EPFL
# Copyright 2025 Open Brain Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from enum import Enum
from typing import Optional
import logging
from bluecellulab.stimulus.stimulus import DelayedZap, Empty, Ramp, Slope, Step, StepNoise, Stimulus, OrnsteinUhlenbeck, ShotNoise, Sinusoidal, Pulse

logger = logging.getLogger(__name__)


class APWaveformTimings(Enum):
    """APWaveform timings."""
    PRE_DELAY = 250.0
    DURATION = 50.0
    POST_DELAY = 250.0


class IDRestTimings(Enum):
    """IDRest timings."""
    PRE_DELAY = 250.0
    DURATION = 1350.0
    POST_DELAY = 250.0


class IVTimings(Enum):
    """IV timings."""
    PRE_DELAY = 250.0
    DURATION = 3000.0
    POST_DELAY = 250.0


class FirePatternTimings(Enum):
    """FirePattern timings."""
    PRE_DELAY = 250.0
    DURATION = 3600.0
    POST_DELAY = 250.0


class PosCheopsTimings(Enum):
    """PosCheops timings."""
    PRE_DELAY = 250.0
    RAMP1_DURATION = 4000.0
    RAMP2_DURATION = 2000.0
    RAMP3_DURATION = 1333.0
    INTER_DELAY = 2000.0
    POST_DELAY = 250.0


class NegCheopsTimings(Enum):
    """NegCheops timings."""
    PRE_DELAY = 1750.0
    RAMP1_DURATION = 3333.0
    RAMP2_DURATION = 1666.0
    RAMP3_DURATION = 1111.0
    INTER_DELAY = 2000.0
    POST_DELAY = 250.0


class SineSpecTimings(Enum):
    """SineSpec timings."""
    PRE_DELAY = 0
    DURATION = 5000.0
    POST_DELAY = 0


class StimulusFactory:
    def __init__(self, dt: float):
        self.dt = dt

    def step(
        self, pre_delay: float, duration: float, post_delay: float, amplitude: float
    ) -> Stimulus:
        return Step.amplitude_based(
            self.dt,
            pre_delay=pre_delay,
            duration=duration,
            post_delay=post_delay,
            amplitude=amplitude,
        )

    def ramp(
        self,
        pre_delay: float,
        duration: float,
        post_delay: float,
        amplitude: Optional[float] = None,
        threshold_current: Optional[float] = None,
        threshold_percentage: Optional[float] = 220.0,
    ) -> Stimulus:
        if amplitude is not None:
            if threshold_current is not None and threshold_current != 0 and threshold_percentage is not None:
                logger.info(
                    "amplitude, threshold_current and threshold_percentage are all set in ramp."
                    " Will only keep amplitude value."
                )
            return Ramp.amplitude_based(
                self.dt,
                pre_delay=pre_delay,
                duration=duration,
                post_delay=post_delay,
                amplitude=amplitude,
            )

        if threshold_current is not None and threshold_current != 0 and threshold_percentage is not None:
            return Ramp.threshold_based(
                self.dt,
                pre_delay=pre_delay,
                duration=duration,
                post_delay=post_delay,
                threshold_current=threshold_current,
                threshold_percentage=threshold_percentage,
            )

        raise TypeError("You have to give either threshold_current or amplitude")

    def ap_waveform(
        self,
        threshold_current: Optional[float] = None,
        threshold_percentage: Optional[float] = 220.0,
        amplitude: Optional[float] = None,
    ) -> Stimulus:
        """Returns the APWaveform Stimulus object, a type of Step stimulus.

        Args:
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
            amplitude: Raw amplitude of input current.
        """

        if amplitude is not None:
            if threshold_current is not None and threshold_current != 0 and threshold_percentage is not None:
                logger.info(
                    "amplitude, threshold_current and threshold_percentage are all set in ap_waveform."
                    " Will only keep amplitude value."
                )
            return Step.amplitude_based(
                self.dt,
                pre_delay=APWaveformTimings.PRE_DELAY.value,
                duration=APWaveformTimings.DURATION.value,
                post_delay=APWaveformTimings.POST_DELAY.value,
                amplitude=amplitude,
            )

        if threshold_current is not None and threshold_current != 0 and threshold_percentage is not None:
            return Step.threshold_based(
                self.dt,
                pre_delay=APWaveformTimings.PRE_DELAY.value,
                duration=APWaveformTimings.DURATION.value,
                post_delay=APWaveformTimings.POST_DELAY.value,
                threshold_current=threshold_current,
                threshold_percentage=threshold_percentage,
            )

        raise TypeError("You have to give either threshold_current or amplitude")

    def idrest(
        self,
        threshold_current: Optional[float] = None,
        threshold_percentage: Optional[float] = 200.0,
        amplitude: Optional[float] = None,
    ) -> Stimulus:
        """Returns the IDRest Stimulus object, a type of Step stimulus.

        Args:
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
            amplitude: Raw amplitude of input current.
        """

        if amplitude is not None:
            if threshold_current is not None and threshold_current != 0 and threshold_percentage is not None:
                logger.info(
                    "amplitude, threshold_current and threshold_percentage are all set in idrest."
                    " Will only keep amplitude value."
                )
            return Step.amplitude_based(
                self.dt,
                pre_delay=IDRestTimings.PRE_DELAY.value,
                duration=IDRestTimings.DURATION.value,
                post_delay=IDRestTimings.POST_DELAY.value,
                amplitude=amplitude,
            )

        if threshold_current is not None and threshold_current != 0 and threshold_percentage is not None:
            return Step.threshold_based(
                self.dt,
                pre_delay=IDRestTimings.PRE_DELAY.value,
                duration=IDRestTimings.DURATION.value,
                post_delay=IDRestTimings.POST_DELAY.value,
                threshold_current=threshold_current,
                threshold_percentage=threshold_percentage,
            )

        raise TypeError("You have to give either threshold_current or amplitude")

    def iv(
        self,
        threshold_current: Optional[float] = None,
        threshold_percentage: Optional[float] = -40.0,
        amplitude: Optional[float] = None,
    ) -> Stimulus:
        """Returns the IV Stimulus object, a type of Step stimulus.

        Args:
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
            amplitude: Raw amplitude of input current.
        """

        if amplitude is not None:
            if threshold_current is not None and threshold_current != 0 and threshold_percentage is not None:
                logger.info(
                    "amplitude, threshold_current and threshold_percentage are all set in iv."
                    " Will only keep amplitude value."
                )
            return Step.amplitude_based(
                self.dt,
                pre_delay=IVTimings.PRE_DELAY.value,
                duration=IVTimings.DURATION.value,
                post_delay=IVTimings.POST_DELAY.value,
                amplitude=amplitude,
            )

        if threshold_current is not None and threshold_current != 0 and threshold_percentage is not None:
            return Step.threshold_based(
                self.dt,
                pre_delay=IVTimings.PRE_DELAY.value,
                duration=IVTimings.DURATION.value,
                post_delay=IVTimings.POST_DELAY.value,
                threshold_current=threshold_current,
                threshold_percentage=threshold_percentage,
            )

        raise TypeError("You have to give either threshold_current or amplitude")

    def fire_pattern(
        self,
        threshold_current: Optional[float] = None,
        threshold_percentage: Optional[float] = 200.0,
        amplitude: Optional[float] = None,
    ) -> Stimulus:
        """Returns the FirePattern Stimulus object, a type of Step stimulus.

        Args:
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
            amplitude: Raw amplitude of input current.
        """

        if amplitude is not None:
            if threshold_current is not None and threshold_current != 0 and threshold_percentage is not None:
                logger.info(
                    "amplitude, threshold_current and threshold_percentage are all set in fire_pattern."
                    " Will only keep amplitude value."
                )
            return Step.amplitude_based(
                self.dt,
                pre_delay=FirePatternTimings.PRE_DELAY.value,
                duration=FirePatternTimings.DURATION.value,
                post_delay=FirePatternTimings.POST_DELAY.value,
                amplitude=amplitude,
            )

        if threshold_current is not None and threshold_current != 0 and threshold_percentage is not None:
            return Step.threshold_based(
                self.dt,
                pre_delay=FirePatternTimings.PRE_DELAY.value,
                duration=FirePatternTimings.DURATION.value,
                post_delay=FirePatternTimings.POST_DELAY.value,
                threshold_current=threshold_current,
                threshold_percentage=threshold_percentage,
            )

        raise TypeError("You have to give either threshold_current or amplitude")

    def pos_cheops(
        self,
        threshold_current: Optional[float] = None,
        threshold_percentage: Optional[float] = 300.0,
        amplitude: Optional[float] = None,
    ) -> Stimulus:
        """A combination of pyramid shaped Ramp stimuli with a positive
        amplitude.

        Args:
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
            amplitude: Raw amplitude of input current.
        """

        if amplitude is None:
            if threshold_current is None or threshold_current == 0 or threshold_percentage is None:
                raise TypeError("You have to give either threshold_current or amplitude")
            amplitude = threshold_current * threshold_percentage / 100
        elif threshold_current is not None and threshold_current != 0 and threshold_percentage is not None:
            logger.info(
                "amplitude, threshold_current and threshold_percentage are all set in pos_cheops."
                " Will only keep amplitude value."
            )
        result = (
            Empty(self.dt, duration=PosCheopsTimings.PRE_DELAY.value)
            + Slope(self.dt, duration=PosCheopsTimings.RAMP1_DURATION.value, amplitude_start=0.0, amplitude_end=amplitude)
            + Slope(self.dt, duration=PosCheopsTimings.RAMP1_DURATION.value, amplitude_start=amplitude, amplitude_end=0.0)
            + Empty(self.dt, duration=PosCheopsTimings.INTER_DELAY.value)
            + Slope(self.dt, duration=PosCheopsTimings.RAMP2_DURATION.value, amplitude_start=0.0, amplitude_end=amplitude)
            + Slope(self.dt, duration=PosCheopsTimings.RAMP2_DURATION.value, amplitude_start=amplitude, amplitude_end=0.0)
            + Empty(self.dt, duration=PosCheopsTimings.INTER_DELAY.value)
            + Slope(self.dt, duration=PosCheopsTimings.RAMP3_DURATION.value, amplitude_start=0.0, amplitude_end=amplitude)
            + Slope(self.dt, duration=PosCheopsTimings.RAMP3_DURATION.value, amplitude_start=amplitude, amplitude_end=0.0)
            + Empty(self.dt, duration=PosCheopsTimings.POST_DELAY.value)
        )
        return result

    def neg_cheops(
        self,
        threshold_current: Optional[float] = None,
        threshold_percentage: Optional[float] = 300.0,
        amplitude: Optional[float] = None,
    ) -> Stimulus:
        """A combination of pyramid shaped Ramp stimuli with a negative
        amplitude.

        Args:
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
            amplitude: Raw amplitude of input current.
        """

        if amplitude is None:
            if threshold_current is None or threshold_current == 0 or threshold_percentage is None:
                raise TypeError("You have to give either threshold_current or amplitude")
            amplitude = - threshold_current * threshold_percentage / 100
        elif threshold_current is not None and threshold_current != 0 and threshold_percentage is not None:
            logger.info(
                "amplitude, threshold_current and threshold_percentage are all set in neg_cheops."
                " Will only keep amplitude value."
            )
        result = (
            Empty(self.dt, duration=NegCheopsTimings.PRE_DELAY.value)
            + Slope(self.dt, duration=NegCheopsTimings.RAMP1_DURATION.value, amplitude_start=0.0, amplitude_end=amplitude)
            + Slope(self.dt, duration=NegCheopsTimings.RAMP1_DURATION.value, amplitude_start=amplitude, amplitude_end=0.0)
            + Empty(self.dt, duration=NegCheopsTimings.INTER_DELAY.value)
            + Slope(self.dt, duration=NegCheopsTimings.RAMP2_DURATION.value, amplitude_start=0.0, amplitude_end=amplitude)
            + Slope(self.dt, duration=NegCheopsTimings.RAMP2_DURATION.value, amplitude_start=amplitude, amplitude_end=0.0)
            + Empty(self.dt, duration=NegCheopsTimings.INTER_DELAY.value)
            + Slope(self.dt, duration=NegCheopsTimings.RAMP3_DURATION.value, amplitude_start=0.0, amplitude_end=amplitude)
            + Slope(self.dt, duration=NegCheopsTimings.RAMP3_DURATION.value, amplitude_start=amplitude, amplitude_end=0.0)
            + Empty(self.dt, duration=NegCheopsTimings.POST_DELAY.value)
        )
        return result

    def sinespec(
        self,
        threshold_current: Optional[float] = None,
        threshold_percentage: Optional[float] = 60.0,
        amplitude: Optional[float] = None,
        pre_delay: float = 0,
    ) -> Stimulus:
        """Returns the SineSpec Stimulus object, a type of Zap stimulus.

        Args:
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
            amplitude: Raw amplitude of input current.
            pre_delay: delay before the start of the stimulus
        """

        if amplitude is not None:
            if threshold_current is not None and threshold_current != 0 and threshold_percentage is not None:
                logger.info(
                    "amplitude, threshold_current and threshold_percentage are all set in sinespec."
                    " Will only keep amplitude value."
                )
            return DelayedZap.amplitude_based(
                self.dt,
                pre_delay=pre_delay,
                duration=SineSpecTimings.DURATION.value,
                post_delay=SineSpecTimings.POST_DELAY.value,
                amplitude=amplitude,
            )

        if threshold_current is not None and threshold_current != 0 and threshold_percentage is not None:
            return DelayedZap.threshold_based(
                self.dt,
                pre_delay=pre_delay,
                duration=SineSpecTimings.DURATION.value,
                post_delay=SineSpecTimings.POST_DELAY.value,
                threshold_current=threshold_current,
                threshold_percentage=threshold_percentage,
            )

        raise TypeError("You have to give either threshold_current or amplitude")

    def sinusoidal(
        self,
        pre_delay: float,
        post_delay: float,
        duration: float,
        frequency: float,
        amplitude: Optional[float] = None,
        amplitude_percent: Optional[float] = None,
        threshold_current: Optional[float] = None,
        dt: float = 0.025,
    ) -> Stimulus:
        """Creates a Sinusoidal stimulus (factory-compatible).

        Args:
            pre_delay: Delay before the sinusoidal stimulus starts (ms).
            post_delay: Delay after the stimulus ends (ms).
            duration: Duration of the stimulus (ms).
            frequency: Frequency of oscillation (Hz).
            amplitude: Absolute amplitude (nA). Used if provided.
            amplitude_percent: Amplitude as a percentage of threshold current.
            threshold_current: Reference threshold current for percentage-based calculation.
            dt: Time step of the stimulus (ms).

        Returns:
            A `Stimulus` object (Sinusoidal) that can be plotted and injected.

        Notes:
            - If `amplitude` is provided, `amplitude_percent` is ignored.
            - If `threshold_current` is not provided, threshold-based parameters cannot be used.
        """
        is_amplitude_based = amplitude is not None
        is_threshold_based = (
            threshold_current is not None
            and threshold_current != 0
            and amplitude_percent is not None
        )

        if is_amplitude_based:
            if is_threshold_based:
                logger.info(
                    "amplitude, threshold_current, and amplitude_percent are all set in Sinusoidal."
                    " Using absolute amplitude and ignoring threshold-based parameters."
                )

            return Sinusoidal.amplitude_based(
                dt=dt,
                pre_delay=pre_delay,
                post_delay=post_delay,
                duration=duration,
                frequency=frequency,
                amplitude=amplitude,  # type: ignore[arg-type]
            )

        if is_threshold_based:
            return Sinusoidal.threshold_based(
                dt=dt,
                pre_delay=pre_delay,
                post_delay=post_delay,
                duration=duration,
                frequency=frequency,
                amplitude_percent=amplitude_percent,  # type: ignore[arg-type]
                threshold_current=threshold_current,  # type: ignore[arg-type]
            )

        raise TypeError("You have to provide either `amplitude` or `threshold_current` with `amplitude_percent`.")

    def ornstein_uhlenbeck(
        self,
        pre_delay: float,
        post_delay: float,
        duration: float,
        tau: float,
        sigma: Optional[float] = None,
        mean: Optional[float] = None,
        mean_percent: Optional[float] = None,
        sigma_percent: Optional[float] = None,
        threshold_current: Optional[float] = None,
        seed: Optional[int] = None
    ) -> Stimulus:
        """Creates an Ornstein-Uhlenbeck process stimulus (factory-compatible).

        Args:
            pre_delay: Delay before the noise starts (ms).
            post_delay: Delay after the noise ends (ms).
            duration: Duration of the stimulus (ms).
            tau: Time constant of the noise process.
            sigma: Standard deviation of the noise (used when mean is provided).
            mean: Absolute mean current value (used if provided).
            mean_percent: Mean current as a percentage of threshold current (used if mean is None).
            sigma_percent: Standard deviation as a percentage of threshold current (used if sigma is None).
            threshold_current: Reference threshold current for percentage-based calculation.
            seed: Optional random seed for reproducibility.

        Returns:
            A `Stimulus` object (OrnsteinUhlenbeck) that can be plotted and injected.

        Notes:
            - If `mean` is provided, `mean_percent` is ignored.
            - If `threshold_current` is not provided, threshold-based parameters cannot be used.
        """
        is_amplitude_based = mean is not None and sigma is not None
        is_threshold_based = (
            threshold_current is not None
            and threshold_current != 0
            and mean_percent is not None
            and sigma_percent is not None
        )

        if is_amplitude_based:
            if is_threshold_based:
                logger.info(
                    "mean, threshold_current, and mean_percent are all set in Ornstein-Uhlenbeck."
                    " Using mean and ignoring threshold-based parameters."
                )

            return OrnsteinUhlenbeck.amplitude_based(
                dt=self.dt,
                pre_delay=pre_delay,
                post_delay=post_delay,
                duration=duration,
                tau=tau,
                sigma=sigma,  # type: ignore[arg-type]
                mean=mean,  # type: ignore[arg-type]
                seed=seed
            )

        if is_threshold_based:
            return OrnsteinUhlenbeck.threshold_based(
                dt=self.dt,
                pre_delay=pre_delay,
                post_delay=post_delay,
                duration=duration,
                mean_percent=mean_percent,  # type: ignore[arg-type]
                sigma_percent=sigma_percent,  # type: ignore[arg-type]
                threshold_current=threshold_current,  # type: ignore[arg-type]
                tau=tau,
                seed=seed
            )

        raise TypeError("You have to give either `mean` and `sigma` or `threshold_current` and `mean_percent` and `sigma_percent`.")

    def shot_noise(
        self,
        pre_delay: float,
        post_delay: float,
        duration: float,
        rate: float,
        rise_time: float,
        decay_time: float,
        mean: Optional[float] = None,
        sigma: Optional[float] = None,
        mean_percent: Optional[float] = None,
        sigma_percent: Optional[float] = None,
        relative_skew: float = 0.5,
        threshold_current: Optional[float] = None,
        seed: Optional[int] = None
    ) -> Stimulus:
        """Creates a ShotNoise instance, either with an absolute amplitude or
        relative to a threshold current.

        Args:
            pre_delay: Delay before the noise starts (ms).
            post_delay: Delay after the noise ends (ms).
            duration: Duration of the stimulus (ms).
            rate: Mean rate of synaptic events (Hz).
            mean: Mean amplitude of events (nA), used if provided.
            sigma: Standard deviation of event amplitudes.
            rise_time: Rise time of synaptic events (ms).
            decay_time: Decay time of synaptic events (ms).
            mean_percent: Mean current as a percentage of threshold current (used if mean is None).
            sigma_percent: Standard deviation as a percentage of threshold current (used if sigma is None).
            relative_skew: Skew factor for the shot noise process (default: 0.5).
            threshold_current: Reference threshold current for percentage-based calculation.
            seed: Optional random seed for reproducibility.

        Returns:
            A `Stimulus` object that can be plotted and injected.

        Notes:
            - If `mean` is provided, `mean_percent` is ignored.
            - If `threshold_current` is not provided, threshold-based parameters cannot be used.
        """
        is_amplitude_based = mean is not None and sigma is not None
        is_threshold_based = (
            threshold_current is not None
            and threshold_current != 0
            and mean_percent is not None
            and sigma_percent is not None
        )

        if is_amplitude_based:
            if is_threshold_based:
                logger.info(
                    "mean, threshold_current, and mean_percent are all set in ShotNoise."
                    " Using mean and ignoring threshold-based parameters."
                )

            return ShotNoise.amplitude_based(
                dt=self.dt,
                pre_delay=pre_delay,
                post_delay=post_delay,
                duration=duration,
                rate=rate,
                mean=mean,  # type: ignore[arg-type]
                sigma=sigma,  # type: ignore[arg-type]
                rise_time=rise_time,
                decay_time=decay_time,
                seed=seed
            )

        if is_threshold_based:
            return ShotNoise.threshold_based(
                dt=self.dt,
                pre_delay=pre_delay,
                post_delay=post_delay,
                duration=duration,
                rise_time=rise_time,
                decay_time=decay_time,
                mean_percent=mean_percent,  # type: ignore[arg-type]
                sigma_percent=sigma_percent,  # type: ignore[arg-type]
                threshold_current=threshold_current,  # type: ignore[arg-type]
                relative_skew=relative_skew,
                seed=seed
            )

        raise TypeError("You must provide either `mean` and `sigma`, or `threshold_current` and `mean_percent` and `sigma_percent` with percentage values.")

    def step_noise(
        self,
        pre_delay: float,
        post_delay: float,
        duration: float,
        step_interval: float,
        mean: Optional[float] = None,
        sigma: Optional[float] = None,
        mean_percent: Optional[float] = None,
        sigma_percent: Optional[float] = None,
        threshold_current: Optional[float] = None,
        seed: Optional[int] = None,
    ) -> Stimulus:
        """Creates a StepNoise instance, either with an absolute amplitude or
        relative to a threshold current.

        Args:
            pre_delay: Delay before the step noise starts (ms).
            post_delay: Delay after the step noise ends (ms).
            duration: Duration of the stimulus (ms).
            step_interval: Interval at which noise amplitude changes.
            mean: Mean amplitude of step noise (nA), used if provided.
            sigma: Standard deviation of step noise.
            mean_percent: Mean current as a percentage of threshold current (used if mean is None).
            sigma_percent: Standard deviation as a percentage of threshold current (used if sigma is None).
            threshold_current: Reference threshold current for percentage-based calculation.
            seed: Optional random seed for reproducibility.

        Returns:
            A `Stimulus` object that can be plotted and injected.

        Notes:
            - If `mean` is provided, `mean_percent` is ignored.
            - If `threshold_current` is not provided, threshold-based parameters cannot be used.
        """
        is_amplitude_based = mean is not None and sigma is not None
        is_threshold_based = (
            threshold_current is not None
            and threshold_current != 0
            and mean_percent is not None
            and sigma_percent is not None
        )

        if is_amplitude_based:
            if is_threshold_based:
                logger.info(
                    "mean, threshold_current, and mean_percent are all set in StepNoise."
                    " Using mean and ignoring threshold-based parameters."
                )
            return StepNoise.amplitude_based(
                dt=self.dt,
                pre_delay=pre_delay,
                post_delay=post_delay,
                duration=duration,
                step_interval=step_interval,
                mean=mean,  # type: ignore[arg-type]
                sigma=sigma,  # type: ignore[arg-type]
                seed=seed,
            )

        if is_threshold_based:
            return StepNoise.threshold_based(
                dt=self.dt,
                pre_delay=pre_delay,
                post_delay=post_delay,
                duration=duration,
                step_interval=step_interval,
                mean_percent=mean_percent,  # type: ignore[arg-type]
                sigma_percent=sigma_percent,  # type: ignore[arg-type]
                threshold_current=threshold_current,  # type: ignore[arg-type]
                seed=seed,
            )

        raise TypeError("You must provide either `mean` and `sigma`, or `threshold_current` and `mean_percent` and `sigma_percent`  with percentage values.")

    def pulse(
        self,
        pre_delay: float,
        duration: float,
        post_delay: float,
        frequency: float,
        width: float,
        threshold_current: Optional[float] = None,
        threshold_percentage: Optional[float] = None,
        amplitude: Optional[float] = None,
    ) -> Stimulus:
        """Creates a pulse stimulus.

        Args:
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
            amplitude: Raw amplitude of input current.
            pre_delay: Delay before the first pulse (ms).
            duration: Duration of the pulse train (ms).
            post_delay: Delay after the last pulse (ms).
            frequency: Frequency of the pulses (Hz).
            width: Width of each pulse (ms).
        """
        if amplitude is not None:
            if threshold_current is not None and threshold_current != 0 and threshold_percentage is not None:
                logger.info(
                    "amplitude, threshold_current and threshold_percentage are all set in pulse."
                    " Will only keep amplitude value."
                )
            return Pulse.amplitude_based(self.dt, pre_delay, duration, post_delay, amplitude, frequency, width)

        if threshold_current is not None and threshold_current != 0 and threshold_percentage is not None:
            return Pulse.threshold_based(self.dt, pre_delay, duration, post_delay, threshold_current, threshold_percentage, frequency, width)

        raise TypeError("You have to give either threshold_current or amplitude")



# Copyright 2023-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Representing the circuit format."""

from __future__ import annotations
from enum import Enum
import json
from pathlib import Path

from bluecellulab.circuit.config import BluepySimulationConfig, SimulationConfig, SonataSimulationConfig


class CircuitFormat(Enum):
    SONATA = "sonata"
    BLUECONFIG = "blueconfig"


def determine_circuit_format(circuit: str | Path | SimulationConfig) -> CircuitFormat:
    """Returns the circuit format for the input circuit path."""
    if isinstance(circuit, SonataSimulationConfig):
        return CircuitFormat.SONATA
    elif isinstance(circuit, BluepySimulationConfig):
        return CircuitFormat.BLUECONFIG
    else:  # not possibly a SimulationConfig
        if is_valid_json_file(circuit):  # type: ignore
            return CircuitFormat.SONATA
        else:
            return CircuitFormat.BLUECONFIG


def is_valid_json_file(fpath: str | Path) -> bool:
    """Check if the input file is a valid json."""
    try:
        json.load(open(fpath))
    except (TypeError, json.JSONDecodeError):
        return False
    return True



# Copyright 2023-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains probability distributions."""

from math import log, sqrt

import neuron


from bluecellulab.type_aliases import NeuronRNG, NeuronVector


# Gamma-distributed sample generator (not available in NEURON)
def gamma(rng: NeuronRNG, a: float, b: float, N: int = 1) -> NeuronVector:
    """Sample N variates from a gamma distribution with parameters shape = a,
    scale = b using the NEURON random number generator rng.

    Uses the algorithm by Marsaglia and Tsang 2001.
    """
    if a < 1:
        rng.uniform(0, 1)
        w = neuron.h.Vector(N)
        w.setrand(rng)
        w.pow(1 / a)
        return gamma(rng, 1 + a, b, N).mul(w)

    d = a - 1 / 3
    c = 1 / 3 / sqrt(d)

    vec = neuron.h.Vector(N)
    for i in range(0, N):
        while True:
            x = rng.normal(0, 1)
            v = 1 + c * x
            if v > 0:
                v = v * v * v
                u = rng.uniform(0, 1)
                if u < 1 - 0.0331 * x * x * x * x:
                    vec.x[i] = b * d * v
                    break
                if log(u) < 0.5 * x * x + d * (1 - v + log(v)):
                    vec.x[i] = b * d * v
                    break

    return vec



# Copyright 2023-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module to interpret NEURON code strings."""

from __future__ import annotations
import ast
import sys
from typing import Any

from bluecellulab.exceptions import NeuronEvalError

PY39_PLUS = sys.version_info >= (3, 9)


def _recursive_evaluate(node: ast.AST, context: dict[str, Any]) -> Any:
    """A limited evaluator for evaluating NEURON code string."""
    if isinstance(node, ast.Constant):
        return node.value
    if isinstance(node, ast.Name):
        return context[node.id]
    if isinstance(node, ast.Attribute):
        base = _recursive_evaluate(node.value, context)
        return getattr(base, node.attr)
    if isinstance(node, ast.Call):
        func = _recursive_evaluate(node.func, context)
        args = [_recursive_evaluate(arg, context) for arg in node.args]
        return func(*args)
    if isinstance(node, ast.Subscript):
        base = _recursive_evaluate(node.value, context)
        if PY39_PLUS:
            index = _recursive_evaluate(node.slice, context)
        else:
            index = _recursive_evaluate(node.slice.value, context)  # type: ignore
        return base[index]
    raise NeuronEvalError("Unexpected code!")


def eval_neuron(source: str, **context) -> Any:
    """A limited interpreter for evaluating NEURON code."""
    tree = ast.parse(source)

    if len(tree.body) != 1:
        raise NeuronEvalError("NEURON code should be a single expression")

    [node] = tree.body
    if not isinstance(node, ast.Expr):
        raise NeuronEvalError("NEURON code should be an expression")

    return _recursive_evaluate(node.value, context)



"""Module for analyzing cell simulation results."""
try:
    import efel
except ImportError:
    efel = None
from itertools import islice
import logging
from matplotlib.collections import LineCollection
import matplotlib.pyplot as plt
import neuron
import numpy as np
import pathlib
import seaborn as sns

from bluecellulab import Cell
from bluecellulab.analysis.inject_sequence import run_stimulus
from bluecellulab.analysis.plotting import plot_iv_curve, plot_fi_curve
from bluecellulab.analysis.utils import exp_decay
from bluecellulab.simulation import Simulation
from bluecellulab.stimulus import StimulusFactory
from bluecellulab.stimulus.circuit_stimulus_definitions import Hyperpolarizing
from bluecellulab.tools import calculate_rheobase


logger = logging.getLogger(__name__)


def compute_plot_iv_curve(cell,
                          injecting_section="soma[0]",
                          injecting_segment=0.5,
                          recording_section="soma[0]",
                          recording_segment=0.5,
                          stim_start=100.0,
                          duration=500.0,
                          post_delay=100.0,
                          threshold_voltage=-20,
                          nb_bins=11,
                          rheobase=None,
                          show_figure=True,
                          save_figure=False,
                          output_dir="./",
                          output_fname="iv_curve.pdf"):
    """Compute and plot the Current-Voltage (I-V) curve for a given cell by
    injecting a range of currents.

    This function evaluates the relationship between the injected current amplitude and the resulting
    steady-state membrane potential of a neuronal cell model. Currents are injected at a specified section
    and segment, and the steady-state voltage at the recording location is used to construct the I-V curve.

    Args:
        cell (bluecellulab.cell.Cell): The initialized BlueCelluLab cell model.
        injecting_section (str, optional): The name of the section where the stimulus is injected.
            Default is "soma[0]".
        injecting_segment (float, optional): The position along the injecting section (0.0 to 1.0)
            where the stimulus is applied. Default is 0.5.
        recording_section (str, optional): The name of the section where the voltage is recorded.
            Default is "soma[0]".
        recording_segment (float, optional): The position along the recording section (0.0 to 1.0)
            where the voltage is recorded. Default is 0.5.
        stim_start (float, optional): The start time of the current injection (in ms). Default is 100.0 ms.
        duration (float, optional): The duration of the current injection (in ms). Default is 500.0 ms.
        post_delay (float, optional): The delay after the stimulation ends before the simulation stops
            (in ms). Default is 100.0 ms.
        threshold_voltage (float, optional): The voltage threshold (in mV) for detecting a steady-state
            response. Default is -20 mV.
        nb_bins (int, optional): The number of discrete current levels between 0 and the maximum current.
            Default is 11.
        rheobase (float, optional): The rheobase current (in nA) for the cell. If not provided, it will
            be calculated using the `calculate_rheobase` function.
        show_figure (bool): Whether to display the figure. Default is True.
        save_figure (bool): Whether to save the figure. Default is False.
        output_dir (str): The directory to save the figure if save_figure is True. Default is "./".
        output_fname (str): The filename to save the figure as if save_figure is True. Default is "iv_curve.png".

    Returns:
        tuple: A tuple containing:
            - list_amp (np.ndarray): The injected current amplitudes (nA).
            - steady_states (np.ndarray): The corresponding steady-state voltages (mV) recorded at the
              specified location.

    Raises:
        ValueError: If the cell object is invalid, the specified sections/segments are not found, or if
            the simulation results are inconsistent.
    """
    if rheobase is None:
        rheobase = calculate_rheobase(cell=cell, section=injecting_section, segx=injecting_segment)

    list_amp = np.linspace(rheobase - 2, rheobase - 0.1, nb_bins)  # [nA]

    steps = []
    times = []
    voltages = []
    # inject step current and record voltage response
    stim_factory = StimulusFactory(dt=0.1)
    for amp in list_amp:
        step_stimulus = stim_factory.step(pre_delay=stim_start, duration=duration, post_delay=post_delay, amplitude=amp)
        recording = run_stimulus(cell.template_params,
                                 step_stimulus,
                                 section=injecting_section,
                                 segment=injecting_segment,
                                 recording_section=recording_section,
                                 recording_segment=recording_segment)
        steps.append(step_stimulus)
        times.append(recording.time)
        voltages.append(recording.voltage)

    steady_states = []
    # compute steady state response
    efel.set_setting('Threshold', threshold_voltage)
    for voltage, t in zip(voltages, times):
        trace = {
            'T': t,
            'V': voltage,
            'stim_start': [stim_start],
            'stim_end': [stim_start + duration]
        }
        features_results = efel.get_feature_values([trace], ['steady_state_voltage_stimend'])
        steady_state = features_results[0]['steady_state_voltage_stimend'][0]
        steady_states.append(steady_state)

    plot_iv_curve(list_amp,
                  steady_states,
                  injecting_section=injecting_section,
                  injecting_segment=injecting_segment,
                  recording_section=recording_section,
                  recording_segment=recording_segment,
                  show_figure=show_figure,
                  save_figure=save_figure,
                  output_dir=output_dir,
                  output_fname=output_fname)

    return np.array(list_amp), np.array(steady_states)


def compute_plot_fi_curve(cell,
                          injecting_section="soma[0]",
                          injecting_segment=0.5,
                          recording_section="soma[0]",
                          recording_segment=0.5,
                          stim_start=100.0,
                          duration=500.0,
                          post_delay=100.0,
                          max_current=0.8,
                          threshold_voltage=-20,
                          nb_bins=11,
                          rheobase=None,
                          show_figure=True,
                          save_figure=False,
                          output_dir="./",
                          output_fname="fi_curve.pdf"):
    """Compute and plot the Frequency-Current (F-I) curve for a given cell by
    injecting a range of currents.

    This function evaluates the relationship between injected current amplitude and the firing rate
    of a neuronal cell model. Currents are injected at a specified section and segment, and the number
    of spikes recorded in the specified recording location is used to construct the F-I curve.

    Args:
        cell (bluecellulab.cell.Cell): The initialized BlueCelluLab cell model.
        injecting_section (str, optional): The name of the section where the stimulus is injected.
            Default is "soma[0]".
        injecting_segment (float, optional): The position along the injecting section (0.0 to 1.0)
            where the stimulus is applied. Default is 0.5.
        recording_section (str, optional): The name of the section where spikes are recorded.
            Default is "soma[0]".
        recording_segment (float, optional): The position along the recording section (0.0 to 1.0)
            where spikes are recorded. Default is 0.5.
        stim_start (float, optional): The start time of the current injection (in ms). Default is 100.0 ms.
        duration (float, optional): The duration of the current injection (in ms). Default is 500.0 ms.
        post_delay (float, optional): The delay after the stimulation ends before the simulation stops
            (in ms). Default is 100.0 ms.
        max_current (float, optional): The maximum amplitude of the injected current (in nA).
            Default is 0.8 nA.
        threshold_voltage (float, optional): The voltage threshold (in mV) for detecting a steady-state
            response. Default is -20 mV.
        nb_bins (int, optional): The number of discrete current levels between 0 and `max_current`.
            Default is 11.
        rheobase (float, optional): The rheobase current (in nA) for the cell. If not provided, it will
            be calculated using the `calculate_rheobase` function.
        show_figure (bool): Whether to display the figure. Default is True.
        save_figure (bool): Whether to save the figure. Default is False.
        output_dir (str): The directory to save the figure if save_figure is True. Default is "./".
        output_fname (str): The filename to save the figure as if save_figure is True. Default is "iv_curve.png".

    Returns:
        tuple: A tuple containing:
            - list_amp (np.ndarray): The injected current amplitudes (nA).
            - spike_count (np.ndarray): The corresponding spike counts for each current amplitude.

    Raises:
        ValueError: If the cell object is invalid or the specified sections/segments are not found.
    """
    if rheobase is None:
        rheobase = calculate_rheobase(cell=cell, section=injecting_section, segx=injecting_segment)

    list_amp = np.linspace(rheobase, max_current, nb_bins)  # [nA]
    steps = []
    spikes = []
    # inject step current and record spike response
    stim_factory = StimulusFactory(dt=0.1)
    for amp in list_amp:
        step_stimulus = stim_factory.step(pre_delay=stim_start, duration=duration, post_delay=post_delay, amplitude=amp)
        recording = run_stimulus(cell.template_params,
                                 step_stimulus,
                                 section=injecting_section,
                                 segment=injecting_segment,
                                 recording_section=recording_section,
                                 recording_segment=recording_segment,
                                 enable_spike_detection=True,
                                 threshold_spike_detection=threshold_voltage)
        steps.append(step_stimulus)
        spikes.append(recording.spike)

    spike_count = [len(spike) for spike in spikes]

    plot_fi_curve(list_amp,
                  spike_count,
                  injecting_section=injecting_section,
                  injecting_segment=injecting_segment,
                  recording_section=recording_section,
                  recording_segment=recording_segment,
                  show_figure=show_figure,
                  save_figure=save_figure,
                  output_dir=output_dir,
                  output_fname=output_fname)

    return np.array(list_amp), np.array(spike_count)


class BPAP:
    # taken from the examples

    def __init__(self, cell: Cell) -> None:
        self.cell = cell
        self.dt = 0.025
        self.stim_start = 1000
        self.stim_duration = 3
        self.basal_cmap = sns.color_palette("crest", as_cmap=True)
        self.apical_cmap = sns.color_palette("YlOrBr_r", as_cmap=True)

    @property
    def start_index(self) -> int:
        """Get the index of the start of the stimulus."""
        return int(self.stim_start / self.dt)

    @property
    def end_index(self) -> int:
        """Get the index of the end of the stimulus."""
        return int((self.stim_start + self.stim_duration) / self.dt)

    def get_recordings(self):
        """Get the soma, basal and apical recordings."""
        all_recordings = self.cell.get_allsections_voltagerecordings()
        soma_rec = None
        dend_rec = {}
        apic_rec = {}
        for key, value in all_recordings.items():
            if "soma" in key:
                soma_rec = value
            elif "dend" in key:
                dend_rec[key] = value
            elif "apic" in key:
                apic_rec[key] = value

        return soma_rec, dend_rec, apic_rec

    def run(self, duration: float, amplitude: float) -> None:
        """Apply depolarization and hyperpolarization at the same time."""
        sim = Simulation()
        sim.add_cell(self.cell)
        self.cell.add_allsections_voltagerecordings()
        self.cell.add_step(start_time=self.stim_start, stop_time=self.stim_start + self.stim_duration, level=amplitude)
        hyperpolarizing = Hyperpolarizing("single-cell", delay=0, duration=duration)
        self.cell.add_replay_hypamp(hyperpolarizing)
        sim.run(duration, dt=self.dt, cvode=False)

    def amplitudes(self, recs) -> list[float]:
        """Return amplitude across given sections."""
        efel_feature_name = "maximum_voltage_from_voltagebase"
        traces = [
            {
                'T': self.cell.get_time(),
                'V': rec,
                'stim_start': [self.stim_start],
                'stim_end': [self.stim_start + self.stim_duration]
            }
            for rec in recs.values()
        ]
        features_results = efel.get_feature_values(traces, [efel_feature_name])
        amps = [feat_res[efel_feature_name][0] for feat_res in features_results]

        return amps

    def distances_to_soma(self, recs) -> list[float]:
        """Return the distance to the soma for each section."""
        res = []
        soma = self.cell.soma
        for key in recs.keys():
            section_name = key.rsplit(".")[-1].split("[")[0]  # e.g. "dend"
            section_idx = int(key.rsplit(".")[-1].split("[")[1].split("]")[0])  # e.g. 0
            attribute_value = getattr(self.cell.cell.getCell(), section_name)
            section = next(islice(attribute_value, section_idx, None))
            # section e.g. cADpyr_L2TPC_bluecellulab_x[0].dend[0]
            res.append(neuron.h.distance(soma(0.5), section(0.5)))
        return res

    def get_amplitudes_and_distances(self):
        soma_rec, dend_rec, apic_rec = self.get_recordings()
        soma_amp = self.amplitudes({"soma": soma_rec})[0]
        dend_amps = None
        dend_dist = None
        apic_amps = None
        apic_dist = None
        if dend_rec:
            dend_amps = self.amplitudes(dend_rec)
            dend_dist = self.distances_to_soma(dend_rec)
        if apic_rec:
            apic_amps = self.amplitudes(apic_rec)
            apic_dist = self.distances_to_soma(apic_rec)

        return soma_amp, dend_amps, dend_dist, apic_amps, apic_dist

    def fit(self, soma_amp, dend_amps, dend_dist, apic_amps, apic_dist):
        """Fit the amplitudes vs distances to an exponential decay function."""
        from scipy.optimize import curve_fit

        popt_dend = None
        if dend_amps and dend_dist:
            dist = [0] + dend_dist  # add soma distance
            amps = [soma_amp] + dend_amps  # add soma amplitude
            popt_dend, _ = curve_fit(exp_decay, dist, amps)

        popt_apic = None
        if apic_amps and apic_dist:
            dist = [0] + apic_dist  # add soma distance
            amps = [soma_amp] + apic_amps  # add soma amplitude
            popt_apic, _ = curve_fit(exp_decay, dist, amps)

        return popt_dend, popt_apic

    def validate(self, soma_amp, dend_amps, dend_dist, apic_amps, apic_dist):
        """Check that the exponential fit is decaying."""
        validated = True
        notes = ""
        popt_dend, popt_apic = self.fit(soma_amp, dend_amps, dend_dist, apic_amps, apic_dist)
        if popt_dend is None:
            logger.debug("No dendritic recordings found.")
            notes += "No dendritic recordings found.\n"
        elif popt_dend[1] <= 0:
            logger.debug("Dendritic fit is not decaying.")
            validated = False
            notes += "Dendritic fit is not decaying.\n"
        else:
            notes += "Dendritic validation passed: dendritic amplitude is decaying with distance relative to soma.\n"
        if popt_apic is None:
            logger.debug("No apical recordings found.")
            notes += "No apical recordings found.\n"
        elif popt_apic[1] <= 0:
            logger.debug("Apical fit is not decaying.")
            validated = False
            notes += "Apical fit is not decaying.\n"
        else:
            notes += "Apical validation passed: apical amplitude is decaying with distance relative to soma.\n"

        return validated, notes

    def plot_amp_vs_dist(
        self,
        soma_amp,
        dend_amps,
        dend_dist,
        apic_amps,
        apic_dist,
        show_figure=True,
        save_figure=False,
        output_dir="./",
        output_fname="bpap.pdf",
    ):
        """Plot the results of the BPAP analysis."""
        popt_dend, popt_apic = self.fit(soma_amp, dend_amps, dend_dist, apic_amps, apic_dist)

        outpath = pathlib.Path(output_dir) / output_fname
        fig, ax1 = plt.subplots(figsize=(10, 6))
        ax1.scatter([0], [soma_amp], marker="^", color='black', label='Soma')
        if dend_amps and dend_dist:
            ax1.scatter(
                dend_dist,
                dend_amps,
                c=dend_dist,
                cmap=self.basal_cmap,
                label='Basal Dendrites',
            )
            if popt_dend is not None:
                x = np.linspace(0, max(dend_dist), 100)
                y = exp_decay(x, *popt_dend)
                ax1.plot(x, y, color='darkgreen', linestyle='--', label='Basal Dendritic Fit')
        if apic_amps and apic_dist:
            ax1.scatter(
                apic_dist,
                apic_amps,
                c=apic_dist,
                cmap=self.apical_cmap,
                label='Apical Dendrites'
            )
            if popt_apic is not None:
                x = np.linspace(0, max(apic_dist), 100)
                y = exp_decay(x, *popt_apic)
                ax1.plot(x, y, color='goldenrod', linestyle='--', label='Apical Fit')
        ax1.set_xlabel('Distance to Soma (um)')
        ax1.set_ylabel('Amplitude (mV)')
        ax1.legend()
        fig.suptitle('Back-propagating Action Potential Analysis')
        fig.tight_layout()
        if save_figure:
            fig.savefig(outpath)
        if show_figure:
            plt.show()

        return outpath

    def plot_one_axis_recordings(self, fig, ax, rec_list, dist, cmap):
        """Plot the soma and dendritic recordings on one axis.

        Args:
            fig (matplotlib.figure.Figure): The figure to plot on.
            ax (matplotlib.axes.Axes): The axis to plot on.
            rec_list (list): List of recordings to plot.
            dist (list): List of distances from the soma for each recording.
            cmap (matplotlib.colors.Colormap): Colormap to use for the recordings.
        """
        time = self.cell.get_time()
        line_collection = LineCollection(
            [np.column_stack([time, rec]) for rec in rec_list],
            array=dist,
            cmap=cmap,
        )
        ax.set_xlim(
            self.stim_start - 0.1,
            self.stim_start + 30
        )
        ax.set_ylim(
            min([min(rec[self.start_index:]) for rec in rec_list]) - 2,
            max([max(rec[self.start_index:]) for rec in rec_list]) + 2
        )
        ax.add_collection(line_collection)
        fig.colorbar(line_collection, label="soma distance (um)", ax=ax)

    def plot_recordings(
        self,
        show_figure=True,
        save_figure=False,
        output_dir="./",
        output_fname="bpap_recordings.pdf",
    ):
        """Plot the recordings from all dendrites."""
        soma_rec, dend_rec, apic_rec = self.get_recordings()
        dend_dist = []
        apic_dist = []
        if dend_rec:
            dend_dist = self.distances_to_soma(dend_rec)
        if apic_rec:
            apic_dist = self.distances_to_soma(apic_rec)
        # add soma_rec to the lists
        dend_rec_list = [soma_rec] + list(dend_rec.values())
        dend_dist = [0] + dend_dist
        apic_rec_list = [soma_rec] + list(apic_rec.values())
        apic_dist = [0] + apic_dist

        outpath = pathlib.Path(output_dir) / output_fname
        fig, (ax1, ax2) = plt.subplots(figsize=(10, 12), nrows=2, sharex=True)

        self.plot_one_axis_recordings(fig, ax1, dend_rec_list, dend_dist, self.basal_cmap)
        self.plot_one_axis_recordings(fig, ax2, apic_rec_list, apic_dist, self.apical_cmap)

        # plt.setp(ax1.get_xticklabels(), visible=False)
        ax1.set_title('Basal Dendritic Recordings')
        ax2.set_title('Apical Dendritic Recordings')
        ax1.set_ylabel('Voltage (mV)')
        ax2.set_ylabel('Voltage (mV)')
        ax2.set_xlabel('Time (ms)')
        fig.suptitle('Back-propagating Action Potential Recordings')
        fig.tight_layout()
        if save_figure:
            fig.savefig(outpath)
        if show_figure:
            plt.show()

        return outpath



# Copyright 2023-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Plotting functions for the cell module."""

from __future__ import annotations

import neuron

import bluecellulab


class PlottableMixin:
    """Mixin responsible of plotting functions of a cell."""

    def __init__(self) -> None:
        """Store the persistent objects."""
        self.cell_dendrograms: list[bluecellulab.Dendrogram] = []
        self.plot_windows: list[bluecellulab.PlotWindow] = []

        self.fih_plots = None
        self.fih_weights = None

        # As long as no PlotWindow or active Dendrogram exist, don't update
        self.plot_callback_necessary = False

    def delete_plottable(self) -> None:
        """NEURON state to be cleaned upon object destruction."""
        self.fih_plots = None
        self.fih_weights = None

    def add_plot_window(self, var_list, xlim=None, ylim=None, title=""):
        """Add a window to plot a variable."""
        xlim = [0, 1000] if xlim is None else xlim
        ylim = [-100, 100] if ylim is None else ylim
        for var_name in var_list:
            if var_name not in self.recordings:
                self.add_recording(var_name)
        self.plot_windows.append(bluecellulab.PlotWindow(
            var_list, self, xlim, ylim, title))
        self.plot_callback_necessary = True

    def add_dendrogram(
            self,
            variable=None,
            active=False,
            save_fig_path=None,
            interactive=False,
            scale_bar=True,
            scale_bar_size=10.0,
            fig_title=None):
        """Show a dendrogram of the cell."""
        self._init_psections()
        cell_dendrogram = bluecellulab.Dendrogram(
            self.psections,
            variable=variable,
            active=active,
            save_fig_path=save_fig_path,
            interactive=interactive,
            scale_bar=scale_bar,
            scale_bar_size=scale_bar_size,
            fig_title=fig_title)
        cell_dendrogram.redraw()
        self.cell_dendrograms.append(cell_dendrogram)
        if active:
            self.plot_callback_necessary = True

    def init_callbacks(self):
        """Initialize the callback function (if necessary)."""
        if not self.delayed_weights.empty():
            self.fih_weights = neuron.h.FInitializeHandler(
                1, self.weights_callback)

        if self.plot_callback_necessary:
            self.fih_plots = neuron.h.FInitializeHandler(1, self.plot_callback)

    def weights_callback(self):
        """Callback function that updates the delayed weights, when a certain
        delay has been reached."""
        while not self.delayed_weights.empty() and \
                abs(self.delayed_weights.queue[0][0] - neuron.h.t) < \
                neuron.h.dt:
            (_, (sid, weight)) = self.delayed_weights.get()
            if sid in self.connections:
                if self.connections[sid].post_netcon is not None:
                    self.connections[sid].post_netcon.weight[0] = weight

        if not self.delayed_weights.empty():
            neuron.h.cvode.event(
                self.delayed_weights.queue[0][0], self.weights_callback
            )

    def plot_callback(self):
        """Update all the windows."""
        for window in self.plot_windows:
            window.redraw()
        for cell_dendrogram in self.cell_dendrograms:
            cell_dendrogram.redraw()

        neuron.h.cvode.event(
            neuron.h.t + 1, self.plot_callback)



# Copyright 2023-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Class that represents a connection between two cells in bluecellulab."""

from typing import Optional

import neuron
import numpy as np

from bluecellulab.cell.core import Cell
from bluecellulab.circuit import SynapseProperty


class Connection:
    """Class that represents a connection between two cells in bluecellulab."""

    def __init__(
            self,
            post_synapse,
            pre_spiketrain: Optional[np.ndarray] = None,
            pre_cell: Optional[Cell] = None,
            stim_dt=None,
            parallel_context=None,
            spike_threshold: float = -30.0,
            spike_location="soma"):
        self.persistent = []
        self.delay = post_synapse.syn_description[SynapseProperty.AXONAL_DELAY]
        self.weight = post_synapse.syn_description[SynapseProperty.G_SYNX]
        self.pre_cell = pre_cell
        self.pre_spiketrain = pre_spiketrain
        self.post_synapse = post_synapse
        self.pc = parallel_context

        if post_synapse.weight is not None:
            self.weight_scalar = post_synapse.weight
        else:
            self.weight_scalar = 1.0

        self.post_netcon = None
        self.post_netcon_delay = self.delay
        self.post_netcon_weight = self.weight * self.weight_scalar

        if self.pre_spiketrain is not None:
            if any(self.pre_spiketrain < 0):
                raise ValueError("bluecellulab Connection: a spiketrain contains "
                                 "a negative time, this is not supported by "
                                 "NEURON's Vecstim: %s" %
                                 str(self.pre_spiketrain))
            t_vec = neuron.h.Vector(self.pre_spiketrain)
            vecstim = neuron.h.VecStim()
            vecstim.play(t_vec, stim_dt)
            self.post_netcon = neuron.h.NetCon(
                vecstim, self.post_synapse.hsynapse,
                spike_threshold,
                self.post_netcon_delay,
                self.post_netcon_weight)
            # set netcon type
            nc_param_name = 'nc_type_param_{}'.format(
                self.post_synapse.hsynapse).split('[')[0]
            if hasattr(neuron.h, nc_param_name):
                nc_type_param = int(getattr(neuron.h, nc_param_name))
                self.post_netcon.weight[nc_type_param] = 2  # NC_REPLAY
            self.persistent.append(t_vec)
            self.persistent.append(vecstim)
        elif self.pre_cell is not None:
            self.post_netcon = self.pre_cell.create_netcon_spikedetector(
                self.post_synapse.hsynapse, location=spike_location,
                threshold=spike_threshold) if self.pc is None else \
                self.pc.gid_connect(self.pre_cell.cell_id.id, self.post_synapse.hsynapse)
            self.post_netcon.weight[0] = self.post_netcon_weight
            self.post_netcon.delay = self.post_netcon_delay
            self.post_netcon.threshold = spike_threshold
            # set netcon type
            nc_param_name = 'nc_type_param_{}'.format(
                self.post_synapse.hsynapse).split('[')[0]
            if hasattr(neuron.h, nc_param_name):
                nc_type_param = int(getattr(neuron.h, nc_param_name))
                self.post_netcon.weight[nc_type_param] = 0  # NC_PRESYN

    @property
    def info_dict(self):
        """Return dict that contains information that can restore this conn."""

        connection_dict = {}

        connection_dict['pre_cell_id'] = self.post_synapse.pre_gid
        connection_dict['post_cell_id'] = self.post_synapse.post_cell_id.id
        connection_dict['post_synapse_id'] = self.post_synapse.syn_id.sid

        connection_dict['post_netcon'] = {}
        connection_dict['post_netcon']['weight'] = self.post_netcon_weight
        connection_dict['post_netcon']['delay'] = self.post_netcon_delay

        return connection_dict

    def delete(self):
        """Delete the connection."""
        for persistent_object in self.persistent:
            del persistent_object

    def __del__(self):
        self.delete()



# Copyright 2023-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functionality for validation the simulation configuration."""

import re

import pandas as pd

from bluecellulab import BLUEPY_AVAILABLE
from bluecellulab.circuit.circuit_access import CircuitAccess
from bluecellulab.circuit.synapse_properties import SynapseProperty
from bluecellulab.exceptions import ConfigError, TargetDoesNotExist, ExtraDependencyMissingError


class SimulationValidator:
    """Validates the simulation configuration, should be called before
    simulation."""

    def __init__(self, circuit_access: CircuitAccess) -> None:
        if not BLUEPY_AVAILABLE:
            raise ExtraDependencyMissingError("bluepy")
        self.circuit_access = circuit_access

    def validate(self) -> None:
        self.check_connection_entries()
        self.check_cao_cr_glusynapse_value()
        self.check_spike_location()

    def check_connection_entries(self):
        """Check all connection entries at once."""
        gid_pttrn = re.compile("^a[0-9]+")
        for entry in self.circuit_access.config.connection_entries():
            for target in (entry.source, entry.target):
                if not (self.circuit_access.is_valid_group(target) or gid_pttrn.match(target)):
                    raise TargetDoesNotExist("%s target does not exist" % target)

    def check_cao_cr_glusynapse_value(self):
        """Make sure cao_CR_GluSynapse is equal to ExtracellularCalcium."""
        condition_parameters = self.circuit_access.config.condition_parameters()
        if condition_parameters.extracellular_calcium is not None:
            cao_cr_glusynapse = condition_parameters.extracellular_calcium

            if cao_cr_glusynapse != self.circuit_access.config.extracellular_calcium:
                raise ConfigError("cao_CR_GluSynapse is not equal to ExtracellularCalcium")

    def check_spike_location(self):
        """Allow only accepted spike locations."""
        if 'SpikeLocation' in self.circuit_access.config.impl.Run:
            spike_location = self.circuit_access.config.impl.Run['SpikeLocation']
            if spike_location not in ["soma", "AIS"]:
                raise ConfigError(
                    "Possible options for SpikeLocation are 'soma' and 'AIS'")


def check_nrrp_value(synapses: pd.DataFrame) -> None:
    """Assures the nrrp values fits the conditions.

    Args:
        synapses: synapse description

    Raises:
        ValueError: when NRRP is <= 0
        ValueError: when NRRP cannot ve cast to integer
    """
    # remove nan ones, don't check them they're an artifact of pd Join
    nrrp_series = synapses[~synapses[SynapseProperty.NRRP].isna()][SynapseProperty.NRRP]

    if any(nrrp_series <= 0):
        raise ValueError(
            'Value smaller than 0.0 found for Nrrp: '
            f'in {nrrp_series}.'
        )

    if any(nrrp_series != nrrp_series.astype(int)):
        raise ValueError(
            'Non-integer value for Nrrp found: '
            f'{synapses[SynapseProperty.NRRP]} at synapse {synapses}.'
        )



# Copyright 2025 Open Brain Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import matplotlib.pyplot as plt
import numpy
import pathlib

import efel

from bluecellulab.analysis.analysis import BPAP
from bluecellulab.analysis.analysis import compute_plot_fi_curve
from bluecellulab.analysis.analysis import compute_plot_iv_curve
from bluecellulab.analysis.inject_sequence import run_multirecordings_stimulus
from bluecellulab.analysis.inject_sequence import run_stimulus
from bluecellulab.stimulus.factory import IDRestTimings
from bluecellulab.stimulus.factory import StimulusFactory
from bluecellulab.tools import calculate_input_resistance
from bluecellulab.tools import calculate_rheobase

logger = logging.getLogger(__name__)


def plot_trace(recording, out_dir, fname, title):
    """Plot a trace with inout current given a recording."""
    outpath = out_dir / fname
    fig, ax1 = plt.subplots(figsize=(10, 6))
    plt.plot(recording.time, recording.voltage, color="black")
    current_axis = ax1.twinx()
    current_axis.plot(recording.time, recording.current, color="gray", alpha=0.6)
    current_axis.set_ylabel("Stimulus Current [nA]")
    fig.suptitle(title)
    ax1.set_xlabel("Time [ms]")
    ax1.set_ylabel("Voltage [mV]")
    fig.tight_layout()
    fig.savefig(outpath)

    return outpath


def plot_traces(recordings, out_dir, fname, title, labels=None, xlim=None):
    """Plot a trace with inout current given a recording."""
    outpath = out_dir / fname
    fig, ax1 = plt.subplots(figsize=(10, 6))
    prop_cycle = plt.rcParams["axes.prop_cycle"]
    colors = prop_cycle.by_key()["color"]
    N_colors = len(colors)
    for i, recording in enumerate(recordings):
        if i == 0:
            color = "black"
        else:
            color = colors[(i - 1) % N_colors]
        label = labels[i] if labels is not None else None
        plt.plot(recording.time, recording.voltage, color=color, label=label)
    current_axis = ax1.twinx()
    current_axis.plot(recordings[0].time, recordings[0].current, color="gray", alpha=0.6)
    current_axis.set_ylabel("Stimulus Current [nA]")
    fig.suptitle(title)
    ax1.set_xlabel("Time [ms]")
    ax1.set_ylabel("Voltage [mV]")
    if labels is not None:
        ax1.legend()
    if xlim is not None:
        ax1.set_xlim(xlim)
    fig.tight_layout()
    fig.savefig(outpath)

    return outpath


def spiking_test(cell, rheobase, out_dir, spike_threshold_voltage=-30.):
    """Spiking test: cell should spike."""
    stim_factory = StimulusFactory(dt=1.0)
    step_stimulus = stim_factory.idrest(threshold_current=rheobase, threshold_percentage=200)
    recording = run_stimulus(
        cell.template_params,
        step_stimulus,
        "soma[0]",
        0.5,
        add_hypamp=True,
        enable_spike_detection=True,
        threshold_spike_detection=spike_threshold_voltage,
    )
    passed = recording.spike is not None and len(recording.spike) > 0

    # plotting
    outpath = plot_trace(
        recording,
        out_dir,
        fname="spiking_test.pdf",
        title="Spiking Test - Step at 200% of Rheobase",
    )

    notes = "Validation passed: Spikes detected." if passed else "Validation failed: No spikes detected."
    return {
        "name": "Simulatable Neuron Spiking Validation",
        "passed": passed,
        "validation_details": notes,
        "figures": [outpath],
    }


def depolarization_block_test(cell, rheobase, out_dir):
    """Depolarization block test: no depolarization block should be detected."""
    # Run the stimulus
    stim_factory = StimulusFactory(dt=1.0)
    step_stimulus = stim_factory.idrest(threshold_current=rheobase, threshold_percentage=200)
    recording = run_stimulus(
        cell.template_params,
        step_stimulus,
        "soma[0]",
        0.5,
        add_hypamp=True,
        enable_spike_detection=False,
    )
    # Check for depolarization block
    trace = {
        "T": recording.time,
        "V": recording.voltage,
        "stim_start": [IDRestTimings.PRE_DELAY.value],
        "stim_end": [IDRestTimings.PRE_DELAY.value + IDRestTimings.DURATION.value],
    }
    features_results = efel.get_feature_values([trace], ["depol_block_bool"])
    depol_block = bool(features_results[0]["depol_block_bool"][0])

    # plotting
    outpath = plot_trace(
        recording,
        out_dir,
        fname="depolarization_block_test.pdf",
        title="Depolarization Block Test - Step at 200% of Rheobase",
    )

    notes = "Validation passed: No depolarization block detected." if not depol_block else "Validation failed: Depolarization block detected."
    return {
        "name": "Simulatable Neuron Depolarization Block Validation",
        "passed": not depol_block,
        "validation_details": notes,
        "figures": [outpath],
    }


def bpap_test(cell, rheobase, out_dir="./"):
    """Back-propagating action potential test: exponential fit should decay.

    Args:
        cell (Cell): The cell to test.
        rheobase (float): The rheobase current to use for the test.
        out_dir (str): Directory to save the figure.
    """
    amplitude = 10. * rheobase  # Use 1000% of the rheobase current
    bpap = BPAP(cell)
    bpap.run(duration=1500, amplitude=amplitude)
    soma_amp, dend_amps, dend_dist, apic_amps, apic_dist = bpap.get_amplitudes_and_distances()
    validated, notes = bpap.validate(soma_amp, dend_amps, dend_dist, apic_amps, apic_dist)
    outpath_amp_dist = bpap.plot_amp_vs_dist(
        soma_amp,
        dend_amps,
        dend_dist,
        apic_amps,
        apic_dist,
        show_figure=False,
        save_figure=True,
        output_dir=out_dir,
        output_fname="bpap.pdf"
    )
    outpath_recordings = bpap.plot_recordings(
        show_figure=False,
        save_figure=True,
        output_dir=out_dir,
        output_fname="bpap_recordings.pdf",
    )

    return {
        "name": "Simulatable Neuron Back-propagating Action Potential Validation",
        "validation_details": notes,
        "passed": validated,
        "figures": [outpath_amp_dist, outpath_recordings],
    }


def ais_spiking_test(cell, rheobase, out_dir, spike_threshold_voltage=-30.):
    """AIS spiking test: axon should spike before soma."""
    name = "Simulatable Neuron AIS Spiking Validation"
    # Check that the cell has an axon
    if len(cell.axonal) == 0:
        return {
            "name": name,
            "passed": True,
            "validation_details": "Validation skipped: Cell does not have an axon section.",
            "figures": [],
        }

    # Run the stimulus
    stim_factory = StimulusFactory(dt=1.0)
    step_stimulus = stim_factory.idrest(threshold_current=rheobase, threshold_percentage=200)
    recordings = run_multirecordings_stimulus(
        cell.template_params,
        step_stimulus,
        "soma[0]",
        0.5,
        add_hypamp=True,
        recording_locations=[("axon[0]", 0.5), ("soma[0]", 0.5)],
        enable_spike_detection=True,
        threshold_spike_detection=spike_threshold_voltage,
    )
    axon_recording, soma_recording = recordings

    # plotting
    outpath1 = plot_traces(
        recordings,
        out_dir,
        fname="ais_spiking_test.pdf",
        title="AIS Spiking Test - Step at 200% of Rheobase",
        labels=["axon[0]", "soma[0]"],
    )
    outpath2 = plot_traces(
        recordings,
        out_dir,
        fname="ais_spiking_test_zoomed.pdf",
        title="AIS Spiking Test - Step at 200% of Rheobase (zoomed)",
        labels=["axon[0]", "soma[0]"],
        xlim=(IDRestTimings.PRE_DELAY.value, IDRestTimings.PRE_DELAY.value + 100),
    )

    # Check for spiking
    for recording in recordings:
        if recording.spike is None or len(recording.spike) == 0:
            return {
                "name": name,
                "passed": False,
                "validation_details": "Validation failed: No spikes detected in one or both recordings.",
                "figures": [outpath1, outpath2],
            }

    # Check if axon spike happens before soma spike
    passed = bool(axon_recording.spike[0] < soma_recording.spike[0])
    notes = (
        "Validation passed: Axon spikes before soma."
        if passed
        else "Validation failed: Axon does not spike before soma."
    )
    return {
        "name": name,
        "passed": passed,
        "validation_details": notes,
        "figures": [outpath1, outpath2],
    }


def hyperpolarization_test(cell, rheobase, out_dir):
    """Hyperpolarization test: hyperpolarized voltage should be lower than RMP."""
    name = "Simulatable Neuron Hyperpolarization Validation"
    # Run the stimulus
    stim_factory = StimulusFactory(dt=1.0)
    step_stimulus = stim_factory.iv(threshold_current=rheobase, threshold_percentage=-40)
    recording = run_stimulus(
        cell.template_params,
        step_stimulus,
        "soma[0]",
        0.5,
        add_hypamp=True,
        enable_spike_detection=False,
    )

    # plotting
    outpath = plot_trace(
        recording,
        out_dir,
        fname="hyperpolarization_test.pdf",
        title="Hyperpolarization Test - Step at -40% of Rheobase",
    )

    # Check for hyperpolarization
    trace = {
        "T": recording.time,
        "V": recording.voltage,
        "stim_start": [IDRestTimings.PRE_DELAY.value],
        "stim_end": [IDRestTimings.PRE_DELAY.value + IDRestTimings.DURATION.value],
    }
    features_results = efel.get_feature_values([trace], ["voltage_base", "steady_state_voltage_stimend"])
    rmp = features_results[0]["voltage_base"][0]
    ss_voltage = features_results[0]["steady_state_voltage_stimend"][0]
    if rmp is None or ss_voltage is None:
        return {
            "name": name,
            "passed": False,
            "validation_details": "Validation failed: Could not determine RMP or steady state voltage.",
            "figures": [outpath],
        }
    hyperpol_bool = bool(ss_voltage < rmp)

    notes = (
        f"Validation passed: Hyperpolarized voltage ({ss_voltage:.2f} mV) is lower than RMP ({rmp:.2f} mV)."
        if hyperpol_bool
        else f"Validation failed: Hyperpolarized voltage ({ss_voltage:.2f} mV) is not lower than RMP ({rmp:.2f} mV)."
    )
    return {
        "name": name,
        "passed": hyperpol_bool,
        "validation_details": notes,
        "figures": [outpath],
    }


def rin_test(rin):
    """Rin should have an acceptable biological range (< 1000 MOhm)"""
    passed = bool(rin < 1000)

    notes = (
        f"Validation passed: Input resistance (Rin) = {rin:.2f} MOhm is smaller than 1000 MOhm."
        if passed
        else f"Validation failed: Input resistance (Rin) = {rin:.2f} MOhm is higher than 1000 MOhm, which is not realistic."
    )
    return {
        "name": "Simulatable Neuron Input Resistance Validation",
        "passed": passed,
        "validation_details": notes,
        "figures": [],
    }


def iv_test(cell, rheobase, out_dir, spike_threshold_voltage=-30.):
    """IV curve should have a positive slope."""
    name = "Simulatable Neuron IV Curve Validation"
    amps, steady_states = compute_plot_iv_curve(
        cell,
        rheobase=rheobase,
        threshold_voltage=spike_threshold_voltage,
        nb_bins=5,
        show_figure=False,
        save_figure=True,
        output_dir=out_dir,
        output_fname="iv_curve.pdf")

    outpath = pathlib.Path(out_dir) / "iv_curve.pdf"

    # Check for positive slope
    if len(amps) < 2 or len(steady_states) < 2:
        return {
            "name": name,
            "passed": False,
            "validation_details": "Validation failed: Not enough data points to determine slope.",
            "figures": [outpath],
        }
    slope = numpy.polyfit(amps, steady_states, 1)[0]
    passed = bool(slope > 0)
    notes = (
        f"Validation passed: Slope of IV curve = {slope:.2f} is positive."
        if passed
        else f"Validation failed: Slope of IV curve = {slope:.2f} is not positive."
    )
    return {
        "name": name,
        "validation_details": notes,
        "passed": passed,
        "figures": [outpath],
    }


def fi_test(cell, rheobase, out_dir, spike_threshold_voltage=-30.):
    """FI curve should have a positive slope."""
    name = "Simulatable Neuron FI Curve Validation"
    amps, spike_counts = compute_plot_fi_curve(
        cell,
        rheobase=rheobase,
        threshold_voltage=spike_threshold_voltage,
        nb_bins=5,
        show_figure=False,
        save_figure=True,
        output_dir=out_dir,
        output_fname="fi_curve.pdf")

    outpath = pathlib.Path(out_dir) / "fi_curve.pdf"

    # Check for positive slope
    if len(amps) < 2 or len(spike_counts) < 2:
        return {
            "name": name,
            "passed": False,
            "validation_details": "Validation failed: Not enough data points to determine slope.",
            "figures": [outpath],
        }
    slope = numpy.polyfit(amps, spike_counts, 1)[0]
    passed = bool(slope > 0)
    notes = (
        f"Validation passed: Slope of FI curve = {slope:.2f} is positive."
        if passed
        else f"Validation failed: Slope of FI curve = {slope:.2f} is not positive."
    )
    return {
        "name": name,
        "validation_details": notes,
        "passed": passed,
        "figures": [outpath],
    }


def run_validations(
    cell, cell_name, spike_threshold_voltage=-30, output_dir="./memodel_validation_figures"
):
    """Run all the validations on the cell.

    Args:
        cell (Cell): The cell to validate.
        cell_name (str): The name of the cell, used in the output directory.
        spike_threshold_voltage (float): The voltage threshold for spike detection.
        output_dir (str): The directory to save the validation figures.
    """
    out_dir = pathlib.Path(output_dir) / cell_name
    out_dir.mkdir(parents=True, exist_ok=True)

    # cell = Cell.from_template_parameters(template_params)
    # get me-model properties
    holding_current = cell.hypamp if cell.hypamp else 0.0
    if cell.threshold:
        rheobase = cell.threshold
    else:
        rheobase = calculate_rheobase(
            cell=cell, section="soma[0]", segx=0.5, threshold_voltage=spike_threshold_voltage
        )
    rin = calculate_input_resistance(
        template_path=cell.template_params.template_filepath,
        morphology_path=cell.template_params.morph_filepath,
        template_format=cell.template_params.template_format,
        emodel_properties=cell.template_params.emodel_properties,
    )

    # Validation 1: Spiking Test
    logger.debug("Running spiking test")
    spiking_test_result = spiking_test(cell, rheobase, out_dir, spike_threshold_voltage)

    # Validation 2: Depolarization Block Test
    logger.debug("Running depolarization block test")
    depolarization_block_result = depolarization_block_test(cell, rheobase, out_dir)

    # Validation 3: Backpropagating AP Test
    logger.debug("Running backpropagating AP test")
    bpap_result = bpap_test(cell, rheobase, out_dir)

    # Validation 4: Postsynaptic Potential Test
    # logger.debug("Running postsynaptic potential test")
    # We have to wait for ProbAMPANMDA_EMS to be present in entitycore to implement this test

    # Validation 5: AIS Spiking Test
    logger.debug("Running AIS spiking test")
    ais_spiking_test_result = ais_spiking_test(cell, rheobase, out_dir, spike_threshold_voltage)

    # Validation 6: Hyperpolarization Test
    logger.debug("Running hyperpolarization test")
    hyperpolarization_result = hyperpolarization_test(cell, rheobase, out_dir)

    # Validation 7: Rin Test
    logger.debug("Running Rin test")
    rin_result = rin_test(rin)

    # # Validation 8: IV Test
    logger.debug("Running IV test")
    iv_test_result = iv_test(cell, rheobase, out_dir, spike_threshold_voltage)

    # # Validation 9: FI Test
    logger.debug("Running FI test")
    fi_test_result = fi_test(cell, rheobase, out_dir, spike_threshold_voltage)

    return {
        "memodel_properties": {
            "holding_current": holding_current,
            "rheobase": rheobase,
            "rin": rin,
        },
        "spiking_test": spiking_test_result,
        "depolarization_block_test": depolarization_block_result,
        "bpap_test": bpap_result,
        "ais_spiking_test": ais_spiking_test_result,
        "hyperpolarization_test": hyperpolarization_result,
        "rin_test": rin_result,
        "iv_test": iv_test_result,
        "fi_test": fi_test_result,
    }



# Copyright 2023-2024 Blue Brain Project / EPFL
# Copyright 2025 Open Brain Institute

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Optional
import logging
import matplotlib.pyplot as plt
import neuron
import numpy as np
from bluecellulab.cell.stimuli_generator import get_relative_shotnoise_params
from bluecellulab.exceptions import BluecellulabError

logger = logging.getLogger(__name__)


class Stimulus(ABC):
    def __init__(self, dt: float) -> None:
        self.dt = dt

    @property
    @abstractmethod
    def time(self) -> np.ndarray:
        """Time values of the stimulus."""
        ...

    @property
    @abstractmethod
    def current(self) -> np.ndarray:
        """Current values of the stimulus."""
        ...

    def __len__(self) -> int:
        return len(self.time)

    @property
    def stimulus_time(self) -> float:
        return len(self) * self.dt

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(dt={self.dt})"

    def plot(self, ax=None, **kwargs):
        if ax is None:
            ax = plt.gca()
        ax.plot(self.time, self.current, **kwargs)
        ax.set_xlabel("Time (ms)")
        ax.set_ylabel("Current (nA)")
        return ax

    def __add__(self, other: Stimulus) -> CombinedStimulus:
        """Override + operator to concatenate Stimulus objects."""
        if self.dt != other.dt:
            raise ValueError(
                "Stimulus objects must have the same dt to be concatenated"
            )
        if len(self.time) == 0:
            return CombinedStimulus(other.dt, other.time, other.current)
        elif len(other.time) == 0:
            return CombinedStimulus(self.dt, self.time, self.current)
        else:
            # shift other time
            other_time = other.time + self.time[-1] + self.dt
            combined_time = np.concatenate([self.time, other_time])
            # Concatenate the current arrays
            combined_current = np.concatenate([self.current, other.current])
            return CombinedStimulus(self.dt, combined_time, combined_current)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Stimulus):
            return NotImplemented
        else:
            return (
                np.allclose(self.time, other.time)
                and np.allclose(self.current, other.current)
                and self.dt == other.dt
            )


class CombinedStimulus(Stimulus):
    """Represents the Stimulus created by combining multiple stimuli."""

    def __init__(self, dt: float, time: np.ndarray, current: np.ndarray) -> None:
        super().__init__(dt)
        self._time = time
        self._current = current

    @property
    def time(self) -> np.ndarray:
        return self._time

    @property
    def current(self) -> np.ndarray:
        return self._current


class Empty(Stimulus):
    """Represents empty stimulus (all zeros) that has no impact on the cell.

    This is required by some Stimuli that expect the cell to rest.
    """

    def __init__(self, dt: float, duration: float) -> None:
        super().__init__(dt)
        self.duration = duration

    @property
    def time(self) -> np.ndarray:
        return np.arange(0.0, self.duration, self.dt)

    @property
    def current(self) -> np.ndarray:
        return np.zeros_like(self.time)


class Flat(Stimulus):
    def __init__(self, dt: float, duration: float, amplitude: float) -> None:
        super().__init__(dt)
        self.duration = duration
        self.amplitude = amplitude

    @property
    def time(self) -> np.ndarray:
        return np.arange(0.0, self.duration, self.dt)

    @property
    def current(self) -> np.ndarray:
        return np.full_like(self.time, self.amplitude)


class Slope(Stimulus):
    def __init__(
        self, dt: float, duration: float, amplitude_start: float, amplitude_end: float
    ) -> None:
        super().__init__(dt)
        self.duration = duration
        self.amplitude_start = amplitude_start
        self.amplitude_end = amplitude_end

    @property
    def time(self) -> np.ndarray:
        return np.arange(0.0, self.duration, self.dt)

    @property
    def current(self) -> np.ndarray:
        return np.linspace(self.amplitude_start, self.amplitude_end, len(self.time))


class Zap(Stimulus):
    def __init__(self, dt: float, duration: float, amplitude: float) -> None:
        super().__init__(dt)
        self.duration = duration
        self.amplitude = amplitude

    @property
    def time(self) -> np.ndarray:
        return np.arange(0.0, self.duration, self.dt)

    @property
    def current(self) -> np.ndarray:
        return self.amplitude * np.sin(
            2.0 * np.pi * (1.0 + (1.0 / (5.15 - (self.time - 0.1)))) * (self.time - 0.1)
        )


class OUProcess(Stimulus):
    """Generates an Ornstein-Uhlenbeck noise signal."""

    def __init__(self, dt: float, duration: float, tau: float, sigma: float, mean: float, seed: Optional[int] = None):
        super().__init__(dt)  # Ensure proper Stimulus initialization
        self.duration = duration
        self.tau = tau
        self.sigma = sigma
        self.mean = mean
        self.seed = seed

        # Generate OU noise upon initialization
        self._time, self._current = self._generate_ou_noise()

    @property
    def time(self) -> np.ndarray:
        """Returns the time array for the stimulus duration."""
        return self._time

    @property
    def current(self) -> np.ndarray:
        """Returns the Ornstein-Uhlenbeck noise signal."""
        return self._current

    def _generate_ou_noise(self):
        """Generates an Ornstein-Uhlenbeck noise signal."""
        from bluecellulab.cell.stimuli_generator import gen_ornstein_uhlenbeck
        from bluecellulab.rngsettings import RNGSettings

        rng_settings = RNGSettings.get_instance()
        rng = neuron.h.Random()

        if rng_settings.mode == "Random123":
            seed1, seed2, seed3 = 2997, 291204, self.seed if self.seed else 123
            rng.Random123(seed1, seed2, seed3)
        else:
            raise ValueError("Ornstein-Uhlenbeck stimulus requires Random123 RNG mode.")

        # Generate noise signal
        time, current = gen_ornstein_uhlenbeck(self.tau, self.sigma, self.mean, self.duration, self.dt, rng)
        return time, current


class ShotNoiseProcess(Stimulus):
    """Generates a shot noise signal, modeling discrete synaptic events
    occurring at random intervals."""

    def __init__(
        self, dt: float, duration: float, rate: float, mean: float, sigma: float,
        rise_time: float, decay_time: float, seed: Optional[int] = None
    ):
        super().__init__(dt)
        self.duration = duration
        self.rate = rate
        self.mean = mean
        self.sigma = sigma
        self.rise_time = rise_time
        self.decay_time = decay_time
        self.seed = seed

        # Generate shot noise signal
        self._time, self._current = self._generate_shot_noise()

    @property
    def time(self) -> np.ndarray:
        return self._time

    @property
    def current(self) -> np.ndarray:
        return self._current

    def _generate_shot_noise(self):
        """Generates the shot noise time and current vectors."""
        from bluecellulab.cell.stimuli_generator import gen_shotnoise_signal
        from bluecellulab.rngsettings import RNGSettings

        rng_settings = RNGSettings.get_instance()
        rng = neuron.h.Random()

        if rng_settings.mode == "Random123":
            seed1, seed2, seed3 = 2997, 19216, self.seed if self.seed else 123
            rng.Random123(seed1, seed2, seed3)
        else:
            raise ValueError("Shot noise stimulus requires Random123 RNG mode.")

        variance = self.sigma ** 2
        tvec, svec = gen_shotnoise_signal(
            self.decay_time,
            self.rise_time,
            self.rate,
            self.mean,
            variance,
            self.duration,
            self.dt,
            rng=rng
        )

        return np.array(tvec.to_python()), np.array(svec.to_python())


class StepNoiseProcess(Stimulus):
    """Generates step noise: A step current with noise variations."""

    def __init__(
        self,
        dt: float,
        duration: float,
        step_interval: float,
        mean: float,
        sigma: float,
        seed: Optional[int] = None,
    ):
        super().__init__(dt)
        self.duration = duration
        self.step_interval = step_interval
        self.mean = mean
        self.sigma = sigma
        self.seed = seed

        # Generate step noise signal
        self._time, self._current = self._generate_step_noise()

    @property
    def time(self) -> np.ndarray:
        return self._time

    @property
    def current(self) -> np.ndarray:
        return self._current

    def _generate_step_noise(self):
        """Generates the step noise time and current vectors using NEURON’s
        random generator."""
        from bluecellulab.rngsettings import RNGSettings

        # Get NEURON RNG settings
        rng_settings = RNGSettings.get_instance()
        rng = neuron.h.Random()

        if rng_settings.mode == "Random123":
            seed1, seed2, seed3 = 2997, 19216, self.seed if self.seed else 123
            rng.Random123(seed1, seed2, seed3)
        else:
            raise ValueError("StepNoise stimulus requires Random123 RNG mode.")

        num_steps = int(self.duration / self.step_interval)

        # Generate noise using NEURON's normal distribution function
        amplitudes = [self.mean + rng.normal(0, self.sigma) for _ in range(num_steps)]

        # Construct stimulus
        time_values = []
        current_values = []
        time = 0

        for amp in amplitudes:
            time_values.append(time)
            current_values.append(amp)
            time += self.step_interval

        return np.array(time_values), np.array(current_values)


class SinusoidalWave(Stimulus):
    """Generates a sinusoidal current wave."""

    def __init__(self, dt: float, duration: float, amplitude: float, frequency: float):
        super().__init__(dt)
        self.duration = duration
        self.amplitude = amplitude
        self.frequency = frequency

        self._time, self._current = self._generate_sinusoidal_signal()

    @property
    def time(self) -> np.ndarray:
        return self._time

    @property
    def current(self) -> np.ndarray:
        return self._current

    def _generate_sinusoidal_signal(self):
        """Generate the sinusoidal waveform."""
        tvec = neuron.h.Vector()
        tvec.indgen(0.0, self.duration, self.dt)  # Time points using NEURON

        stim = neuron.h.Vector(len(tvec))
        stim.sin(self.frequency, 0.0, self.dt)  # Generate sinusoidal wave using NEURON
        stim.mul(self.amplitude)  # Scale by amplitude

        return np.array(tvec.to_python()), np.array(stim.to_python())


class PulseTrain(Stimulus):
    """Generates a pulse train signal."""

    def __init__(self, dt: float, duration: float, amplitude: float, frequency: float, width: float):
        super().__init__(dt)
        self.duration = duration
        self.amplitude = amplitude
        self.frequency = frequency
        self.width = width

        self._time, self._current = self._generate_pulse_train()

    @property
    def time(self) -> np.ndarray:
        return self._time

    @property
    def current(self) -> np.ndarray:
        return self._current

    def _generate_pulse_train(self):
        total_duration = self.duration
        time_steps = int(total_duration / self.dt)
        time = np.linspace(0, total_duration, time_steps)
        current = np.zeros_like(time)

        start_time = 0
        while start_time + self.width < self.duration:
            pulse_indices = (time >= start_time) & (time < start_time + self.width)
            current[pulse_indices] = self.amplitude
            start_time += 1000.0 / self.frequency

        return time, current


class Step(Stimulus):

    def __init__(self):
        raise NotImplementedError(
            "This class cannot be instantiated directly. "
            "Please use the class methods 'amplitude_based' "
            "or 'threshold_based' to create objects."
        )

    @classmethod
    def amplitude_based(
        cls,
        dt: float,
        pre_delay: float,
        duration: float,
        post_delay: float,
        amplitude: float,
    ) -> CombinedStimulus:
        """Create a Step stimulus from given time events and amplitude.

        Args:
            dt: The time step of the stimulus.
            pre_delay: The delay before the start of the step.
            duration: The duration of the step.
            post_delay: The time to wait after the end of the step.
            amplitude: The amplitude of the step.
        """
        return (
            Empty(dt, duration=pre_delay)
            + Flat(dt, duration=duration, amplitude=amplitude)
            + Empty(dt, duration=post_delay)
        )

    @classmethod
    def threshold_based(
        cls,
        dt: float,
        pre_delay: float,
        duration: float,
        post_delay: float,
        threshold_current: float,
        threshold_percentage: float,
    ) -> CombinedStimulus:
        """Creates a Step stimulus with respect to the threshold current.

        Args:

            dt: The time step of the stimulus.
            pre_delay: The delay before the start of the step.
            duration: The duration of the step.
            post_delay: The time to wait after the end of the step.
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
        """
        amplitude = threshold_current * threshold_percentage / 100
        res = cls.amplitude_based(
            dt,
            pre_delay=pre_delay,
            duration=duration,
            post_delay=post_delay,
            amplitude=amplitude,
        )
        return res


class Ramp(Stimulus):

    def __init__(self):
        raise NotImplementedError(
            "This class cannot be instantiated directly. "
            "Please use the class methods 'amplitude_based' "
            "or 'threshold_based' to create objects."
        )

    @classmethod
    def amplitude_based(
        cls,
        dt: float,
        pre_delay: float,
        duration: float,
        post_delay: float,
        amplitude: float,
    ) -> CombinedStimulus:
        """Create a Ramp stimulus from given time events and amplitudes.

        Args:
            dt: The time step of the stimulus.
            pre_delay: The delay before the start of the ramp.
            duration: The duration of the ramp.
            post_delay: The time to wait after the end of the ramp.
            amplitude: The final amplitude of the ramp.
        """
        return (
            Empty(dt, duration=pre_delay)
            + Slope(
                dt,
                duration=duration,
                amplitude_start=0.0,
                amplitude_end=amplitude,
            )
            + Empty(dt, duration=post_delay)
        )

    @classmethod
    def threshold_based(
        cls,
        dt: float,
        pre_delay: float,
        duration: float,
        post_delay: float,
        threshold_current: float,
        threshold_percentage: float,
    ) -> CombinedStimulus:
        """Creates a Ramp stimulus with respect to the threshold current.

        Args:

            dt: The time step of the stimulus.
            pre_delay: The delay before the start of the ramp.
            duration: The duration of the ramp.
            post_delay: The time to wait after the end of the ramp.
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
        """
        amplitude = threshold_current * threshold_percentage / 100
        res = cls.amplitude_based(
            dt,
            pre_delay=pre_delay,
            duration=duration,
            post_delay=post_delay,
            amplitude=amplitude,
        )
        return res


class DelayedZap(Stimulus):

    def __init__(self):
        raise NotImplementedError(
            "This class cannot be instantiated directly. "
            "Please use the class methods 'amplitude_based' "
            "or 'threshold_based' to create objects."
        )

    @classmethod
    def amplitude_based(
        cls,
        dt: float,
        pre_delay: float,
        duration: float,
        post_delay: float,
        amplitude: float,
    ) -> CombinedStimulus:
        """Create a DelayedZap stimulus from given time events and amplitude.

        Args:
            dt: The time step of the stimulus.
            pre_delay: The delay before the start of the step.
            duration: The duration of the step.
            post_delay: The time to wait after the end of the step.
            amplitude: The amplitude of the step.
        """
        return (
            Empty(dt, duration=pre_delay)
            + Zap(dt, duration=duration, amplitude=amplitude)
            + Empty(dt, duration=post_delay)
        )

    @classmethod
    def threshold_based(
        cls,
        dt: float,
        pre_delay: float,
        duration: float,
        post_delay: float,
        threshold_current: float,
        threshold_percentage: float,
    ) -> CombinedStimulus:
        """Creates a SineSpec stimulus with respect to the threshold current.

        Args:

            dt: The time step of the stimulus.
            pre_delay: The delay before the start of the step.
            duration: The duration of the step.
            post_delay: The time to wait after the end of the step.
            threshold_current: The threshold current of the Cell.
            threshold_percentage: Percentage of desired threshold_current amplification.
        """
        amplitude = threshold_current * threshold_percentage / 100
        res = cls.amplitude_based(
            dt,
            pre_delay=pre_delay,
            duration=duration,
            post_delay=post_delay,
            amplitude=amplitude,
        )
        return res


class OrnsteinUhlenbeck(Stimulus):
    """Factory-compatible Ornstein-Uhlenbeck noise stimulus."""

    def __init__(self):
        """Prevents direct instantiation of the class."""
        raise NotImplementedError(
            "This class cannot be instantiated directly. "
            "Please use 'amplitude_based' or 'threshold_based' methods."
        )

    @classmethod
    def amplitude_based(
        cls,
        dt: float,
        pre_delay: float,
        duration: float,
        post_delay: float,
        tau: float,
        sigma: float,
        mean: float,
        seed: Optional[int] = None,
    ) -> CombinedStimulus:
        """Create an Ornstein-Uhlenbeck stimulus from given time events and
        amplitude."""
        return (
            Empty(dt, duration=pre_delay)
            + OUProcess(dt, duration, tau, sigma, mean, seed)
            + Empty(dt, duration=post_delay)
        )

    @classmethod
    def threshold_based(
        cls,
        dt: float,
        pre_delay: float,
        duration: float,
        post_delay: float,
        mean_percent: float,
        sigma_percent: float,
        threshold_current: float,
        tau: float,
        seed: Optional[int] = None,
    ) -> CombinedStimulus:
        """Creates an Ornstein-Uhlenbeck stimulus with respect to the threshold
        current."""
        sigma = sigma_percent / 100 * threshold_current
        if sigma <= 0:
            raise BluecellulabError(f"Calculated standard deviation (sigma) must be positive, but got {sigma}. Ensure sigma_percent and threshold_current are both positive.")

        mean = mean_percent / 100 * threshold_current
        if mean < 0 and abs(mean) > 2 * sigma:
            logger.warning("Relative Ornstein-Uhlenbeck signal is mostly zero.")

        return cls.amplitude_based(
            dt,
            pre_delay=pre_delay,
            duration=duration,
            post_delay=post_delay,
            tau=tau,
            sigma=sigma,
            mean=mean,
            seed=seed,
        )


class ShotNoise(Stimulus):
    """Factory-compatible Shot Noise Stimulus."""

    def __init__(self):
        """Prevents direct instantiation."""
        raise NotImplementedError(
            "This class cannot be instantiated directly. "
            "Please use 'amplitude_based' or 'threshold_based' methods."
        )

    @classmethod
    def amplitude_based(
        cls,
        dt: float,
        pre_delay: float,
        duration: float,
        post_delay: float,
        rate: float,
        mean: float,
        sigma: float,
        rise_time: float,
        decay_time: float,
        seed: Optional[int] = None,
    ) -> CombinedStimulus:
        """Creates a shot noise stimulus with a specified amplitude.

        Args:
            dt: Time step of the stimulus.
            pre_delay: Delay before the noise starts.
            duration: Duration of the noise signal.
            post_delay: Delay after the noise ends.
            rate: Frequency of synaptic-like events.
            mean: Mean amplitude of the events.
            sigma: Standard deviation of event amplitudes.
            rise_time: Time constant for the event's rise phase.
            decay_time: Time constant for the event's decay phase.
            seed: Random seed for reproducibility.
        """
        return (
            Empty(dt, duration=pre_delay)
            + ShotNoiseProcess(dt, duration, rate, mean, sigma, rise_time, decay_time, seed)
            + Empty(dt, duration=post_delay)
        )

    @classmethod
    def threshold_based(
        cls,
        dt: float,
        pre_delay: float,
        duration: float,
        post_delay: float,
        rise_time: float,
        decay_time: float,
        mean_percent: float,
        sigma_percent: float,
        threshold_current: float,
        relative_skew: float = 0.5,
        seed: Optional[int] = None,
    ) -> CombinedStimulus:
        """Creates a shot noise stimulus based on a neuron's threshold current.

        Args:
            dt: Time step of the stimulus.
            pre_delay: Delay before the noise starts.
            duration: Duration of the noise signal.
            post_delay: Delay after the noise ends.
            rise_time: Rise time constant of events.
            decay_time: Decay time constant of events.
            mean_percent: Mean value as a percentage of the threshold current.
            sigma_percent: Standard deviation as a percentage of the threshold current.
            threshold_current: Baseline threshold current.
            relative_skew: Skew factor affecting noise distribution.
            seed: Random seed for reproducibility.
        """
        _mean = mean_percent / 100 * threshold_current
        sd = sigma_percent / 100 * threshold_current

        rate, mean, sigma = get_relative_shotnoise_params(
            _mean, sd, decay_time, rise_time, relative_skew
        )

        return cls.amplitude_based(
            dt,
            pre_delay=pre_delay,
            duration=duration,
            post_delay=post_delay,
            rate=rate,
            mean=mean,
            sigma=sigma,
            rise_time=rise_time,
            decay_time=decay_time,
            seed=seed,
        )


class StepNoise(Stimulus):
    """Factory-compatible Step Noise Stimulus."""

    def __init__(self):
        """Prevents direct instantiation."""
        raise NotImplementedError(
            "This class cannot be instantiated directly. "
            "Please use 'amplitude_based' or 'threshold_based' methods."
        )

    @classmethod
    def amplitude_based(
        cls,
        dt: float,
        pre_delay: float,
        duration: float,
        post_delay: float,
        step_interval: float,
        mean: float,
        sigma: float,
        seed: Optional[int] = None,
    ) -> CombinedStimulus:
        """Creates a step noise stimulus with a specified amplitude.

        Args:
            dt: Time step of the stimulus.
            pre_delay: Delay before the step noise starts.
            duration: Duration of the noise signal.
            post_delay: Delay after the step noise ends.
            step_interval: Interval at which noise amplitude changes.
            mean: Mean amplitude of step noise.
            sigma: Standard deviation of step noise.
            seed: Random seed for reproducibility.
        """
        return (
            Empty(dt, duration=pre_delay)
            + StepNoiseProcess(dt, duration, step_interval, mean, sigma, seed)
            + Empty(dt, duration=post_delay)
        )

    @classmethod
    def threshold_based(
        cls,
        dt: float,
        pre_delay: float,
        duration: float,
        post_delay: float,
        step_interval: float,
        mean_percent: float,
        sigma_percent: float,
        threshold_current: float,
        seed: Optional[int] = None,
    ) -> CombinedStimulus:
        """Creates a step noise stimulus relative to the threshold current.

        Args:
            dt: Time step of the stimulus.
            pre_delay: Delay before the step noise starts.
            duration: Duration of the noise signal.
            post_delay: Delay after the step noise ends.
            step_interval: Interval at which noise amplitude changes.
            mean_percent: Mean current as a percentage of threshold current.
            sigma_percent: Standard deviation as a percentage of threshold current.
            threshold_current: Baseline threshold current.
            seed: Random seed for reproducibility.
        """
        mean = mean_percent / 100 * threshold_current
        sigma = sigma_percent / 100 * threshold_current

        return cls.amplitude_based(
            dt,
            pre_delay=pre_delay,
            duration=duration,
            post_delay=post_delay,
            step_interval=step_interval,
            mean=mean,
            sigma=sigma,
            seed=seed,
        )


class Sinusoidal(Stimulus):
    """Factory-compatible Sinusoidal Stimulus."""

    def __init__(self):
        """Prevents direct instantiation."""
        raise NotImplementedError(
            "This class cannot be instantiated directly. "
            "Please use 'amplitude_based' or 'threshold_based' methods."
        )

    @classmethod
    def amplitude_based(
        cls,
        dt: float,
        pre_delay: float,
        duration: float,
        post_delay: float,
        amplitude: float,
        frequency: float,
    ) -> CombinedStimulus:
        """Creates a sinusoidal stimulus with a specified amplitude.

        Args:
            dt: Time step of the stimulus.
            pre_delay: Delay before the sinusoidal wave starts.
            duration: Duration of the sinusoidal signal.
            post_delay: Delay after the wave ends.
            amplitude: Amplitude of the sinusoidal wave.
            frequency: Frequency of the wave in Hz.
        """
        return (
            Empty(dt, duration=pre_delay)
            + SinusoidalWave(dt, duration, amplitude, frequency)
            + Empty(dt, duration=post_delay)
        )

    @classmethod
    def threshold_based(
        cls,
        dt: float,
        pre_delay: float,
        duration: float,
        post_delay: float,
        frequency: float,
        threshold_current: float,
        amplitude_percent: float,
    ) -> CombinedStimulus:
        """Creates a sinusoidal stimulus relative to the threshold current.

        Args:
            dt: Time step of the stimulus.
            pre_delay: Delay before the sinusoidal wave starts.
            duration: Duration of the sinusoidal signal.
            post_delay: Delay after the wave ends.
            frequency: Frequency of the wave in Hz.
            threshold_current: Baseline threshold current.
            amplitude_percent: Amplitude as a percentage of the threshold current.
        """
        amplitude = (amplitude_percent / 100) * threshold_current

        return cls.amplitude_based(
            dt,
            pre_delay=pre_delay,
            duration=duration,
            post_delay=post_delay,
            amplitude=amplitude,
            frequency=frequency,
        )


class Pulse(Stimulus):
    """Factory-compatible Pulse Stimulus."""

    def __init__(self):
        """Prevents direct instantiation of the class."""
        raise NotImplementedError(
            "This class cannot be instantiated directly. "
            "Please use 'amplitude_based' or 'threshold_based' methods."
        )

    @classmethod
    def amplitude_based(
        cls, dt: float, pre_delay: float, duration: float, post_delay: float, amplitude: float, frequency: float, width: float
    ) -> CombinedStimulus:
        """Creates a Pulse stimulus from given time events and amplitude."""
        return (
            Empty(dt, duration=pre_delay)
            + PulseTrain(dt, duration, amplitude, frequency, width)
            + Empty(dt, duration=post_delay)
        )

    @classmethod
    def threshold_based(
        cls, dt: float, pre_delay: float, duration: float, post_delay: float, threshold_current: float, threshold_percentage: float,
        frequency: float, width: float
    ) -> CombinedStimulus:
        """Creates a Pulse stimulus with respect to the threshold current."""
        amplitude = threshold_current * (threshold_percentage / 100)
        return cls.amplitude_based(dt, pre_delay, duration, post_delay, amplitude, frequency, width)



# Copyright 2023-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for calculating certain properties of Neurons."""


from __future__ import annotations
from pathlib import Path
from typing import Optional, Tuple
import logging

import neuron
import numpy as np

import bluecellulab
from bluecellulab.cell import Cell
from bluecellulab.circuit.circuit_access import EmodelProperties
from bluecellulab.exceptions import UnsteadyCellError
from bluecellulab.simulation.parallel import IsolatedProcess
from bluecellulab.utils import CaptureOutput
from bluecellulab.type_aliases import NeuronSection


logger = logging.getLogger(__name__)


def calculate_input_resistance(
    template_path: str | Path,
    morphology_path: str | Path,
    template_format: str,
    emodel_properties: EmodelProperties | None,
    current_delta: float = 0.01,
    section: str = "soma[0]",
    segx: float = 0.5
) -> float:
    """Calculate the input resistance at rest of the cell."""
    rest_voltage = calculate_SS_voltage(
        template_path, morphology_path, template_format, emodel_properties, 0.0,
        section=section, segx=segx
    )
    step_voltage = calculate_SS_voltage(
        template_path,
        morphology_path,
        template_format,
        emodel_properties,
        current_delta,
        section=section,
        segx=segx
    )

    voltage_delta = step_voltage - rest_voltage

    return voltage_delta / current_delta


def calculate_SS_voltage(
    template_path: str | Path,
    morphology_path: str | Path,
    template_format: str,
    emodel_properties: EmodelProperties | None,
    step_level: float,
    check_for_spiking=False,
    spike_threshold=-20.0,
    section: str = "soma[0]",
    segx: float = 0.5
) -> float:
    """Calculate the steady state voltage at a certain current step."""
    with IsolatedProcess() as runner:
        SS_voltage = runner.apply(
            calculate_SS_voltage_subprocess,
            [
                template_path,
                morphology_path,
                template_format,
                emodel_properties,
                step_level,
                check_for_spiking,
                spike_threshold,
                section,
                segx
            ],
        )
    return SS_voltage


def calculate_SS_voltage_subprocess(
    template_path: str | Path,
    morphology_path: str | Path,
    template_format: str,
    emodel_properties: EmodelProperties | None,
    step_level: float,
    check_for_spiking: bool,
    spike_threshold: float,
    section: str = "soma[0]",
    segx: float = 0.5
) -> float:
    """Subprocess wrapper of calculate_SS_voltage.

    This code should be run in a separate process. If check_for_spiking
    is True, this function will return None if the cell spikes from
    100ms to the end of the simulation indicating no steady state was
    reached.
    """
    cell = bluecellulab.Cell(
        template_path=template_path,
        morphology_path=morphology_path,
        template_format=template_format,
        emodel_properties=emodel_properties,
    )

    sec = get_section(cell, section)
    neuron_section = sec
    cell.add_voltage_recording(cell.sections[section], segx)
    cell.add_ramp(500, 5000, step_level, step_level, section=neuron_section, segx=segx)
    simulation = bluecellulab.Simulation()
    simulation.run(1000, cvode=template_accepts_cvode(template_path))
    time = cell.get_time()
    voltage = cell.get_voltage_recording(section=neuron_section, segx=segx)
    SS_voltage = np.mean(voltage[np.where((time < 1000) & (time > 800))])
    cell.delete()

    if check_for_spiking:
        # check for voltage crossings
        if len(np.nonzero(voltage[np.where(time > 100.0)] > spike_threshold)[0]) > 0:
            raise UnsteadyCellError(
                "Cell spikes from 100ms to the end of the simulation."
            )

    return SS_voltage


def holding_current_subprocess(v_hold, enable_ttx, cell_kwargs):
    """Subprocess wrapper of holding_current."""
    cell = bluecellulab.Cell(**cell_kwargs)

    if enable_ttx:
        cell.enable_ttx()

    vclamp = neuron.h.SEClamp(0.5, sec=cell.soma)
    vclamp.rs = 0.01
    vclamp.dur1 = 2000
    vclamp.amp1 = v_hold

    simulation = bluecellulab.Simulation()
    simulation.run(1000, cvode=False)

    i_hold = vclamp.i
    v_control = vclamp.vc

    cell.delete()

    return i_hold, v_control


def holding_current(
    v_hold: float,
    cell_id: int | tuple[str, int],
    circuit_path: str | Path,
    enable_ttx=False,
) -> Tuple[float, float]:
    """Calculate the holding current necessary for a given holding voltage."""
    cell_id = bluecellulab.circuit.node_id.create_cell_id(cell_id)
    circuit_sim = bluecellulab.CircuitSimulation(circuit_path)

    cell_kwargs = circuit_sim.fetch_cell_kwargs(cell_id)
    with IsolatedProcess() as runner:
        i_hold, v_control = runner.apply(
            holding_current_subprocess, [v_hold, enable_ttx, cell_kwargs]
        )

    return i_hold, v_control


def template_accepts_cvode(template_name: str | Path) -> bool:
    """Return True if template_name can be run with cvode."""
    with open(template_name, "r") as template_file:
        template_content = template_file.read()
    if "StochKv" in template_content:
        accepts_cvode = False
    else:
        accepts_cvode = True
    return accepts_cvode


def search_hyp_current(
    template_path: str | Path,
    morphology_path: str | Path,
    template_format: str,
    emodel_properties: Optional[EmodelProperties],
    target_voltage: float,
    min_current: float,
    max_current: float,
) -> float:
    """Search current necessary to bring cell to -85 mV."""
    med_current = min_current + abs(min_current - max_current) / 2
    new_target_voltage = calculate_SS_voltage(
        template_path,
        morphology_path,
        template_format,
        emodel_properties,
        med_current,
    )
    logger.info("Detected voltage: %f" % new_target_voltage)
    if abs(new_target_voltage - target_voltage) < 0.5:
        return med_current
    elif new_target_voltage > target_voltage:
        return search_hyp_current(
            template_path=template_path,
            morphology_path=morphology_path,
            template_format=template_format,
            emodel_properties=emodel_properties,
            target_voltage=target_voltage,
            min_current=min_current,
            max_current=med_current,
        )
    else:  # new_target_voltage < target_voltage:
        return search_hyp_current(
            template_path=template_path,
            morphology_path=morphology_path,
            template_format=template_format,
            emodel_properties=emodel_properties,
            target_voltage=target_voltage,
            min_current=med_current,
            max_current=max_current,
        )


def detect_hyp_current(
    template_path: str | Path,
    morphology_path: str | Path,
    template_format: str,
    emodel_properties: EmodelProperties | None,
    target_voltage: float,
) -> float:
    """Search current necessary to bring cell to -85 mV.

    Compared to using NEURON's SEClamp object, the binary search better
    replicates what experimentalists use
    """
    return search_hyp_current(
        template_path=template_path,
        morphology_path=morphology_path,
        template_format=template_format,
        emodel_properties=emodel_properties,
        target_voltage=target_voltage,
        min_current=-1.0,
        max_current=0.0,
    )


def detect_spike_step(
    template_path: str | Path,
    morphology_path: str | Path,
    template_format: str,
    emodel_properties: EmodelProperties | None,
    hyp_level: float,
    inj_start: float,
    inj_stop: float,
    step_level: float,
    section: str = "soma[0]",
    segx: float = 0.5,
    step_thresh: float = -20.
) -> bool:
    """Detect if there is a spike at a certain step level."""
    with IsolatedProcess() as runner:
        spike_detected = runner.apply(
            detect_spike_step_subprocess,
            [
                template_path,
                morphology_path,
                template_format,
                emodel_properties,
                hyp_level,
                inj_start,
                inj_stop,
                step_level,
                section,
                segx,
                step_thresh
            ],
        )
    return spike_detected


def detect_spike_step_subprocess(
    template_path: str | Path,
    morphology_path: str | Path,
    template_format: str,
    emodel_properties: EmodelProperties | None,
    hyp_level: float,
    inj_start: float,
    inj_stop: float,
    step_level: float,
    section: str = "soma[0]",
    segx: float = 0.5,
    step_thresh: float = -20.
) -> bool:
    """Detect if there is a spike at a certain step level."""
    cell = bluecellulab.Cell(
        template_path=template_path,
        morphology_path=morphology_path,
        template_format=template_format,
        emodel_properties=emodel_properties)

    sec = get_section(cell, section)
    cell.add_voltage_recording(cell.sections[section], segx)

    neuron_section = sec
    cell.add_ramp(0, 5000, hyp_level, hyp_level, section=neuron_section, segx=segx)
    cell.add_ramp(inj_start, inj_stop, step_level, step_level, section=neuron_section, segx=segx)
    simulation = bluecellulab.Simulation()
    simulation.run(int(inj_stop), cvode=template_accepts_cvode(template_path))

    time = cell.get_time()
    voltage = cell.get_voltage_recording(section=neuron_section, segx=segx)
    voltage_step = voltage[np.where((time > inj_start) & (time < inj_stop))]
    spike_detected = detect_spike(voltage_step, step_thresh)

    cell.delete()

    return spike_detected


def detect_spike(voltage: np.ndarray, step_thresh: float = -20.) -> bool:
    """Detect if there is a spike in the voltage trace."""
    if len(voltage) == 0:
        return False
    else:
        return bool(np.max(voltage) > step_thresh)  # bool not np.bool_


def search_threshold_current(
    template_name: str | Path,
    morphology_path: str | Path,
    template_format: str,
    emodel_properties: EmodelProperties | None,
    hyp_level: float,
    inj_start: float,
    inj_stop: float,
    min_current: float,
    max_current: float,
    current_precision: float = 0.01,
    section: str = "soma[0]",
    segx: float = 0.5,
    step_thresh: float = -20.
):
    """Search current necessary to reach threshold."""
    if abs(max_current - min_current) < current_precision:
        return max_current
    med_current = min_current + abs(min_current - max_current) / 2
    logger.info("Med current %d" % med_current)

    spike_detected = detect_spike_step(
        template_name, morphology_path, template_format, emodel_properties,
        hyp_level, inj_start, inj_stop, med_current,
        section=section, segx=segx, step_thresh=step_thresh
    )
    logger.info("Spike threshold detection at: %f nA" % med_current)

    if spike_detected:
        return search_threshold_current(template_name, morphology_path,
                                        template_format, emodel_properties,
                                        hyp_level, inj_start, inj_stop,
                                        min_current, med_current,
                                        current_precision,
                                        section=section, segx=segx,
                                        step_thresh=step_thresh)
    else:
        return search_threshold_current(template_name, morphology_path,
                                        template_format, emodel_properties,
                                        hyp_level, inj_start, inj_stop,
                                        med_current, max_current,
                                        current_precision,
                                        section=section, segx=segx,
                                        step_thresh=step_thresh)


def check_empty_topology() -> bool:
    """Return true if NEURON simulator topology command is empty."""
    with CaptureOutput() as stdout:
        neuron.h.topology()

    return stdout == ['', '']


def calculate_max_thresh_current(cell: Cell,
                                 threshold_voltage: float = -30.0,
                                 section: str = "soma[0]",
                                 segx: float = 0.5) -> float:
    """Calculate the upper bound threshold current.

    Args:
        cell (bluecellulab.cell.Cell): The initialized cell model.
        threshold_voltage (float, optional): Voltage threshold for spike detection. Default is -30.0 mV.
        section (str, optional): The section where current is injected.
        segx (float, optional): Fractional location within the section for current injection.

    Returns:
        float: The upper bound threshold current.
    """
    # Calculate resting membrane potential (rmp)
    rmp = calculate_SS_voltage(
        template_path=cell.template_params.template_filepath,
        morphology_path=cell.template_params.morph_filepath,
        template_format=cell.template_params.template_format,
        emodel_properties=cell.template_params.emodel_properties,
        step_level=0.0,
        section=section,
        segx=segx
    )

    # Calculate input resistance (rin)
    rin = calculate_input_resistance(
        template_path=cell.template_params.template_filepath,
        morphology_path=cell.template_params.morph_filepath,
        template_format=cell.template_params.template_format,
        emodel_properties=cell.template_params.emodel_properties,
        section=section,
        segx=segx
    )

    # Calculate upperbound threshold current
    upperbound_threshold_current = (threshold_voltage - rmp) / rin
    upperbound_threshold_current = np.min([upperbound_threshold_current, 2.0])

    return upperbound_threshold_current


def calculate_rheobase(cell: Cell,
                       threshold_voltage: float = -20.0,
                       threshold_search_stim_start: float = 300.0,
                       threshold_search_stim_stop: float = 1000.0,
                       section: str = "soma[0]",
                       segx: float = 0.5) -> float:
    """Calculate the rheobase by first computing the upper bound threshold
    current.

    Args:
        cell (bluecellulab.cell.Cell): The initialized cell model.
        threshold_voltage (float, optional): Voltage threshold for spike detection. Default is -20.0 mV.
        threshold_search_stim_start (float, optional): Start time for threshold search stimulation (in ms). Default is 300.0 ms.
        threshold_search_stim_stop (float, optional): Stop time for threshold search stimulation (in ms). Default is 1000.0 ms.
        section (str, optional): The section where current is injected.
        segx (float, optional): Fractional location within the section for current injection.

    Returns:
        float: The rheobase current.
    """
    if cell.template_params.emodel_properties is None:
        raise ValueError("emodel_properties cannot be None")

    # Calculate upper bound threshold current
    upperbound_threshold_current = calculate_max_thresh_current(
        cell,
        threshold_voltage,
        section,
        segx
    )

    # Compute rheobase
    rheobase = search_threshold_current(
        template_name=cell.template_params.template_filepath,
        morphology_path=cell.template_params.morph_filepath,
        template_format=cell.template_params.template_format,
        emodel_properties=cell.template_params.emodel_properties,
        hyp_level=cell.template_params.emodel_properties.holding_current,
        inj_start=threshold_search_stim_start,
        inj_stop=threshold_search_stim_stop,
        min_current=cell.template_params.emodel_properties.holding_current,
        max_current=upperbound_threshold_current,
        current_precision=0.005,
        section=section,
        segx=segx,
        step_thresh=threshold_voltage
    )

    return rheobase


def validate_section_and_segment(cell: Cell, section_name: str, segment_position: float):
    """Validate a single section and segment position.

    Args:
        cell: The cell model to validate against.
        section_name: The name of the section to validate (e.g., 'soma', 'axon[1]').
        segment_position: The position within the section (e.g., 0.5 for the middle).

    Raises:
        ValueError: If the section or position is invalid.
    """
    # Validate the section
    if section_name not in cell.sections:
        raise ValueError(f"Section '{section_name}' not found in the cell model.")

    # Validate the segment position
    if not (0.0 <= segment_position <= 1.0):
        raise ValueError(f"Segment position must be between 0.0 and 1.0, got {segment_position}.")


def get_section(cell: Cell, section_name: str) -> NeuronSection:
    """Retrieve a NEURON section from the cell by its name.

    Args:
        cell (Cell): The cell object containing the sections.
        section_name (str): The name of the section to retrieve.

    Returns:
        NeuronSection: The NEURON section corresponding to the provided name.

    Raises:
        ValueError: If the section with the specified name does not exist.
    """
    try:
        return cell.sections[section_name]
    except KeyError:
        raise ValueError(f"Section '{section_name}' not found in the cell.")



"""Module for plotting analysis results of cell simulations."""

import matplotlib.pyplot as plt
import pathlib


def plot_iv_curve(
    currents,
    voltages,
    injecting_section,
    injecting_segment,
    recording_section,
    recording_segment,
    show_figure=True,
    save_figure=False,
    output_dir="./",
    output_fname="iv_curve.pdf",
):
    """Plots the IV curve.

    Args:
        currents (list): The injected current levels (nA).
        voltages (list): The corresponding steady-state voltages (mV).
        injecting_section (str): The section in the cell where the current was injected.
        injecting_segment (float): The segment position (0.0 to 1.0) where the current was injected.
        recording_section (str): The section in the cell where spikes were recorded.
        recording_segment (float): The segment position (0.0 to 1.0) where spikes were recorded.
        show_figure (bool): Whether to display the figure. Default is True.
        save_figure (bool): Whether to save the figure. Default is False.
        output_dir (str): The directory to save the figure if save_figure is True. Default is "./".
        output_fname (str): The filename to save the figure as if save_figure is True. Default is "iv_curve.pdf".

    Raises:
        ValueError: If the lengths of currents and voltages do not match.
    """
    if len(currents) != len(voltages):
        raise ValueError("currents and voltages must have the same length")

    plt.figure(figsize=(10, 6))
    plt.plot(currents, voltages, marker='o', linestyle='-', color='b')
    plt.title("I-V Curve")
    plt.xlabel(f"Injected Current [nA] at {injecting_section}({injecting_segment:.2f})")
    plt.ylabel(f"Steady state voltage [mV] at {recording_section}({recording_segment:.2f})")
    plt.grid(True)
    plt.tight_layout()
    if show_figure:
        plt.show()

    if save_figure:
        pathlib.Path(output_dir).mkdir(parents=True, exist_ok=True)
        plt.savefig(pathlib.Path(output_dir) / output_fname, format='pdf')
        plt.close()


def plot_fi_curve(
    currents,
    spike_count,
    injecting_section,
    injecting_segment,
    recording_section,
    recording_segment,
    show_figure=True,
    save_figure=False,
    output_dir="./",
    output_fname="fi_curve.pdf",
):
    """Plots the F-I (Frequency-Current) curve.

    Args:
        currents (list): The injected current levels (nA).
        spike_count (list): The number of spikes recorded for each current level.
        injecting_section (str): The section in the cell where the current was injected.
        injecting_segment (float): The segment position (0.0 to 1.0) where the current was injected.
        recording_section (str): The section in the cell where spikes were recorded.
        recording_segment (float): The segment position (0.0 to 1.0) where spikes were recorded.
        show_figure (bool): Whether to display the figure. Default is True.
        save_figure (bool): Whether to save the figure. Default is False.
        output_dir (str): The directory to save the figure if save_figure is True. Default is "./".
        output_fname (str): The filename to save the figure as if save_figure is True. Default is "fi_curve.pdf".

    Raises:
        ValueError: If the lengths of currents and spike counts do not match.
    """
    if len(currents) != len(spike_count):
        raise ValueError("currents and spike count must have the same length")

    plt.figure(figsize=(10, 6))
    plt.plot(currents, spike_count, marker='o')
    plt.title("F-I Curve")
    plt.xlabel(f"Injected Current [nA] at {injecting_section}({injecting_segment:.2f})")
    plt.ylabel(f"Spike Count recorded at {recording_section}({recording_segment:.2f})")
    plt.grid(True)
    plt.tight_layout()
    if show_figure:
        plt.show()

    if save_figure:
        pathlib.Path(output_dir).mkdir(parents=True, exist_ok=True)
        plt.savefig(pathlib.Path(output_dir) / output_fname, format='pdf')
        plt.close()



# Copyright 2023-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Distance computing functionality between Neuron sections."""

import neuron
import numpy as np


class EuclideanSectionDistance:
    """Calculate euclidian distance between positions on two sections.

    Parameters
    ----------

    hsection1 : hoc section such as cADpyr_L2TPC_bluecellulab[0].apic[1]
                First section
    hsection2 : hoc section
                Second section
    location1 : float
                range x along hsection1
    location2 : float
                range x along hsection2
    dimensions : string
                 planes to project on, e.g. 'xy'
    """

    def __call__(
            self,
            hsection1,
            hsection2,
            location1: float = 0.5,
            location2: float = 0.5,
            dimensions: str = "xyz",
    ):
        """Computes and returns the distance."""
        xs_interp1, ys_interp1, zs_interp1 = self.grindaway(hsection1)
        xs_interp2, ys_interp2, zs_interp2 = self.grindaway(hsection2)

        x1 = xs_interp1[int(np.floor((len(xs_interp1) - 1) * location1))]
        y1 = ys_interp1[int(np.floor((len(ys_interp1) - 1) * location1))]
        z1 = zs_interp1[int(np.floor((len(zs_interp1) - 1) * location1))]

        x2 = xs_interp2[int(np.floor((len(xs_interp2) - 1) * location2))]
        y2 = ys_interp2[int(np.floor((len(ys_interp2) - 1) * location2))]
        z2 = zs_interp2[int(np.floor((len(zs_interp2) - 1) * location2))]

        distance = 0
        if "x" in dimensions:
            distance += (x1 - x2) ** 2
        if "y" in dimensions:
            distance += (y1 - y2) ** 2
        if "z" in dimensions:
            distance += (z1 - z2) ** 2

        distance = np.sqrt(distance)

        return distance

    @staticmethod
    def grindaway(hsection):
        """Grindaway."""
        # get the data for the section
        n_segments = int(neuron.h.n3d(sec=hsection))
        n_comps = hsection.nseg

        xs = np.zeros(n_segments)
        ys = np.zeros(n_segments)
        zs = np.zeros(n_segments)
        lengths = np.zeros(n_segments)
        for index in range(n_segments):
            xs[index] = neuron.h.x3d(index, sec=hsection)
            ys[index] = neuron.h.y3d(index, sec=hsection)
            zs[index] = neuron.h.z3d(index, sec=hsection)
            lengths[index] = neuron.h.arc3d(index, sec=hsection)

        # to use Vector class's .interpolate()
        # must first scale the independent variable
        # i.e. normalize length along centroid
        lengths /= lengths[-1]

        # initialize the destination "independent" vector
        # range = np.array(n_comps+2)
        comp_range = np.arange(0, n_comps + 2) / n_comps - 1.0 / (2 * n_comps)
        comp_range[0] = 0
        comp_range[-1] = 1

        # length contains the normalized distances of the pt3d points
        # along the centroid of the section.  These are spaced at
        # irregular intervals.
        # range contains the normalized distances of the nodes along the
        # centroid of the section.  These are spaced at regular intervals.
        # Ready to interpolate.

        xs_interp = np.interp(comp_range, lengths, xs)
        ys_interp = np.interp(comp_range, lengths, ys)
        zs_interp = np.interp(comp_range, lengths, zs)

        return xs_interp, ys_interp, zs_interp



# Copyright 2023-2024 Blue Brain Project / EPFL

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations
from typing import Optional
from bluecellulab.exceptions import ExtraDependencyMissingError

from bluecellulab import BLUEPY_AVAILABLE

from functools import lru_cache
from pathlib import Path

if BLUEPY_AVAILABLE:
    from bluepy_configfile.configfile import BlueConfig
    from bluepy.utils import open_utf8

from bluecellulab.circuit.config.sections import Conditions, ConnectionOverrides
from bluecellulab.stimulus.circuit_stimulus_definitions import Stimulus


class BluepySimulationConfig:
    """Bluepy implementation of SimulationConfig protocol."""
    _connection_overrides: list[ConnectionOverrides] = []

    def __init__(self, config: str) -> None:
        """A str or a BlueConfig object are valid."""
        if not BLUEPY_AVAILABLE:
            raise ExtraDependencyMissingError("bluepy")
        if isinstance(config, str):
            if not Path(config).exists():
                raise FileNotFoundError(f"Config file {config} not found.")
            else:
                with open_utf8(config) as f:
                    self.impl = BlueConfig(f)
        else:
            self.impl = config

    def get_all_projection_names(self) -> list[str]:
        unique_names = {proj.name for proj in self.impl.typed_sections('Projection')}
        return list(unique_names)

    def get_all_stimuli_entries(self) -> list[Stimulus]:
        """Get all stimuli entries."""
        result = []
        for entry in self.impl.typed_sections('StimulusInject'):
            # retrieve the stimulus to apply
            stimulus_name: str = entry.Stimulus
            # bluepy magic to add underscore Stimulus underscore
            # stimulus_name
            stimulus = self.impl['Stimulus_%s' % stimulus_name].to_dict()
            stimulus["Target"] = entry.Target
            stimulus = Stimulus.from_blueconfig(stimulus)
            if stimulus:
                result.append(stimulus)
        return result

    @lru_cache(maxsize=1)
    def condition_parameters(self) -> Conditions:
        """Returns parameters of global condition block of the blueconfig."""
        try:
            condition_entries = self.impl.typed_sections('Conditions')[0].to_dict()
        except IndexError:
            return Conditions.init_empty()
        return Conditions.from_blueconfig(condition_entries)

    @lru_cache(maxsize=1)
    def _connection_entries(self) -> list[ConnectionOverrides]:
        result = []
        for conn_entry in self.impl.typed_sections('Connection'):
            # SynapseID is ignored by bluepy thus not part of dict representation
            conn_entry_dict = conn_entry.to_dict()
            if "SynapseID" in conn_entry:
                conn_entry_dict["SynapseID"] = int(conn_entry["SynapseID"])
            result.append(ConnectionOverrides.from_blueconfig(conn_entry_dict))
        return result

    def connection_entries(self) -> list[ConnectionOverrides]:
        return self._connection_entries() + self._connection_overrides

    @property
    def base_seed(self) -> int:
        """Base seed of blueconfig."""
        return int(self.impl.Run['BaseSeed']) if 'BaseSeed' in self.impl.Run else 0

    @property
    def synapse_seed(self) -> int:
        """Synapse seed of blueconfig."""
        return int(self.impl.Run['SynapseSeed']) if 'SynapseSeed' in self.impl.Run else 0

    @property
    def ionchannel_seed(self) -> int:
        """Ion channel seed of blueconfig."""
        return int(self.impl.Run['IonChannelSeed']) if 'IonChannelSeed' in self.impl.Run else 0

    @property
    def stimulus_seed(self) -> int:
        """Stimulus seed of blueconfig."""
        return int(self.impl.Run['StimulusSeed']) if 'StimulusSeed' in self.impl.Run else 0

    @property
    def minis_seed(self) -> int:
        """Minis seed of blueconfig."""
        return int(self.impl.Run['MinisSeed']) if 'MinisSeed' in self.impl.Run else 0

    @property
    def rng_mode(self) -> str:
        """Gets the rng mode defined in simulation."""
        # Ugly, but mimicking neurodamus
        if 'Simulator' in self.impl.Run and self.impl.Run['Simulator'] != 'NEURON':
            return 'Random123'
        elif 'RNGMode' in self.impl.Run:
            return self.impl.Run['RNGMode']
        else:
            return "Compatibility"  # default rng mode

    @property
    def spike_threshold(self) -> float:
        """Get the spike threshold from simulation config."""
        if 'SpikeThreshold' in self.impl.Run:
            return float(self.impl.Run['SpikeThreshold'])
        else:
            return -30.0

    @property
    def spike_location(self) -> str:
        """Get the spike location from simulation config."""
        if 'SpikeLocation' in self.impl.Run:
            return self.impl.Run['SpikeLocation']
        else:
            return "soma"

    @property
    def duration(self) -> Optional[float]:
        """Get the duration of the simulation."""
        if 'Duration' in self.impl.Run:
            return float(self.impl.Run['Duration'])
        else:
            return None

    @property
    def dt(self) -> float:
        return float(self.impl.Run['Dt'])

    @property
    def _soma_report_dt(self) -> float:
        return float(self.impl.Report_soma['Dt'])

    @property
    def forward_skip(self) -> Optional[float]:
        if 'ForwardSkip' in self.impl.Run:
            return float(self.impl.Run['ForwardSkip'])
        return None

    @property
    def celsius(self) -> float:
        if 'Celsius' in self.impl.Run:
            return float(self.impl.Run['Celsius'])
        else:
            return 34.0  # default

    @property
    def v_init(self) -> float:
        if 'V_Init' in self.impl.Run:
            return float(self.impl.Run['V_Init'])
        else:
            return -65.0

    @property
    def output_root_path(self) -> str:
        """Get the output root path."""
        return self.impl.Run['OutputRoot']

    @property
    def extracellular_calcium(self) -> Optional[float]:
        """Get the extracellular calcium value."""
        if 'ExtracellularCalcium' in self.impl.Run:
            return float(self.impl.Run['ExtracellularCalcium'])
        else:
            return None

    def add_connection_override(
        self,
        connection_override: ConnectionOverrides
    ) -> None:
        self._connection_overrides.append(connection_override)
